#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from neutronclient.common import exceptions as neutron_client_exc
from neutronclient.v2_0 import client as clientv20
from oslo_concurrency import lockutils
from oslo_config import cfg


neutron_opts = [
    cfg.StrOpt('neutron_server_url',
               default='http://127.0.0.1:9696',
               help='URL for connecting to neutron'),
    cfg.IntOpt('url_timeout',
               default=30,
               help='Timeout value for connecting to neutron in seconds'),
    cfg.StrOpt('admin_user_id',
               help='User id for connecting to neutron in admin context'),
    cfg.StrOpt('admin_username',
               help='Username for connecting to neutron in admin context'),
    cfg.StrOpt('admin_password',
               help='Password for connecting to neutron in admin context',
               secret=True),
    cfg.StrOpt('admin_tenant_id',
               help='Tenant id for connecting to neutron in admin context'),
    cfg.StrOpt('admin_tenant_name',
               help='Tenant name for connecting to neutron in admin context. '
                    'This option will be ignored if neutron_admin_tenant_id '
                    'is set. Note that with Keystone V3 tenant names are '
                    'only unique within a domain.'),
    # region_name required?
    cfg.StrOpt('region_name',
               help='Region name for connecting to neutron in admin context'),
    cfg.StrOpt('admin_auth_url',
               default='http://localhost:5000/v2.0',
               help='Authorization URL for connecting to neutron in admin '
                    'context'),
    cfg.BoolOpt('api_insecure',
                default=False,
                help='If set, ignore any SSL validation issues'),
    cfg.StrOpt('auth_strategy',
               default='keystone',
               help='Authorization strategy for connecting to '
                    'neutron in admin context'),
    # extension_sync_interval required?
    cfg.IntOpt('extension_sync_interval',
               default=600,
               help='Number of seconds before querying neutron for '
                    'extensions'),
    cfg.StrOpt('ca_certificates_file',
               help='Location of CA certificates file to use for '
                    'neutron client requests.'),
]

CONF = cfg.CONF
CONF.register_opts(neutron_opts, 'neutron')


class AdminTokenStore(object):

    _instance = None

    def __init__(self):
        self.admin_auth_token = None

    @classmethod
    def get(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance


def _get_client(token=None, admin=False):
    params = {
        'endpoint_url': CONF.neutron.neutron_server_url,
        'timeout': CONF.neutron.url_timeout,
        'insecure': CONF.neutron.api_insecure,
        'ca_cert': CONF.neutron.ca_certificates_file,
        'auth_strategy': CONF.neutron.auth_strategy,
        'token': token,
    }

    if admin:
        if CONF.neutron.admin_user_id:
            params['user_id'] = CONF.neutron.admin_user_id
        else:
            params['username'] = CONF.neutron.admin_username
        if CONF.neutron.admin_tenant_id:
            params['tenant_id'] = CONF.neutron.admin_tenant_id
        else:
            params['tenant_name'] = CONF.neutron.admin_tenant_name
        params['password'] = CONF.neutron.admin_password
        params['auth_url'] = CONF.neutron.admin_auth_url
    return clientv20.Client(**params)


class ClientWrapper(clientv20.Client):
    '''A neutron client wrapper class.

       Wraps the callable methods, executes it and updates the token,
       as it might change when expires.
    '''

    def __init__(self, base_client):
        # Expose all attributes from the base_client instance
        self.__dict__ = base_client.__dict__
        self.base_client = base_client

    def __getattribute__(self, name):
        obj = object.__getattribute__(self, name)
        if callable(obj):
            obj = object.__getattribute__(self, 'proxy')(obj)
        return obj

    def proxy(self, obj):
        def wrapper(*args, **kwargs):
            ret = obj(*args, **kwargs)
            new_token = self.base_client.get_auth_info()['auth_token']
            _update_token(new_token)
            return ret
        return wrapper


def _update_token(new_token):
    with lockutils.lock('neutron_admin_auth_token_lock'):
        token_store = AdminTokenStore.get()
        token_store.admin_auth_token = new_token


def get_client(context, admin=False):
    if admin or (context.is_admin and not context.auth_token):
        with lockutils.lock('neutron_admin_auth_token_lock'):
            orig_token = AdminTokenStore.get().admin_auth_token
        client = _get_client(orig_token, admin=True)
        return ClientWrapper(client)

    # We got a user token that we can use as-is
    if context.auth_token:
        token = context.auth_token
        return _get_client(token=token)

    # We did not get a user token and we should not be using
    # an admin token so log an error
    raise neutron_client_exc.Unauthorized()



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpclient.v2_0 import client as gbp_client
from keystoneclient import auth as ksc_auth
from keystoneclient import session as ksc_session
from keystoneclient.v3 import client as ksc_client
from neutron._i18n import _LW
from oslo_config import cfg
from oslo_log import log as logging

LOG = logging.getLogger(__name__)

# REVISIT(rkukura): We use keystone to get the name of the keystone
# project owning each neutron resource, which by default, requires
# admin. If we keep this, we should probably move it to a separate
# config module. But we should also investigate whether admin is even
# needed, or if neutron's credentials could somehow be used.
AUTH_GROUP = 'apic_aim_auth'
ksc_session.Session.register_conf_options(cfg.CONF, AUTH_GROUP)
ksc_auth.register_conf_options(cfg.CONF, AUTH_GROUP)


class ProjectNameCache(object):
    """Cache of Keystone project ID to project name mappings."""

    def __init__(self):
        self.project_names = {}
        self.keystone = None
        self.gbp = None

    def _get_keystone_client(self):
        LOG.debug("Getting keystone client")
        auth = ksc_auth.load_from_conf_options(cfg.CONF, AUTH_GROUP)
        LOG.debug("Got auth: %s" % auth)
        if not auth:
            LOG.warning(_LW('No auth_plugin configured in %s'),
                        AUTH_GROUP)
        session = ksc_session.Session.load_from_conf_options(
            cfg.CONF, AUTH_GROUP, auth=auth)
        LOG.debug("Got session: %s" % session)
        self.keystone = ksc_client.Client(session=session)
        LOG.debug("Got keystone client: %s" % self.keystone)
        self.gbp = gbp_client.Client(session=session)
        LOG.debug("Got gbp client: %s" % self.gbp)

    def ensure_project(self, project_id):
        """Ensure cache contains mapping for project.

        :param project_id: ID of the project

        Ensure that the cache contains a mapping for the project
        identified by project_id. If it is not, Keystone will be
        queried for the current list of projects, and any new mappings
        will be added to the cache. This method should never be called
        inside a transaction with a project_id not already in the
        cache.
        """

        if not project_id:
            return

        # TODO(rkukura): It seems load_from_conf_options() and
        # keystoneclient auth plugins have been deprecated, and we
        # should use keystoneauth instead.
        if project_id not in self.project_names:
            if self.keystone is None:
                self._get_keystone_client()
            LOG.debug("Calling project API")
            projects = self.keystone.projects.list()
            LOG.debug("Received projects: %s" % projects)
            for project in projects:
                self.project_names[project.id] = project.name

    def get_project_name(self, project_id):
        """Get name of project from cache.

        :param project_id: ID of the project

        Get the name of the project identified by project_id from the
        cache. If the cache contains project_id, the project's name is
        returned. If not, None is returned.
        """
        return self.project_names.get(project_id)

    def update_project_name(self, project_id):
        if self.keystone is None:
            self._get_keystone_client()
        if self.keystone:
            LOG.debug("Calling project API")
            project = self.keystone.projects.get(project_id)
            # only return project name when there is a change
            if project and self.project_names.get(project_id) != project.name:
                self.project_names[project.id] = project.name
                return project.name
        return None

    def purge_gbp(self, project_id):
        if self.gbp is None:
            self._get_keystone_client()
        if self.gbp:
            LOG.debug("Calling gbp purge() API")
            self.gbp.purge(project_id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import socket
import time

from gbpservice.contrib.nfp.config_orchestrator.common import topics
from gbpservice.nfp.core import log as nfp_logging
from neutron.common import rpc as n_rpc
from neutron.db import agents_db
from neutron.db import agentschedulers_db
from neutron import manager
from neutron_lib import exceptions
from neutron_vpnaas.db.vpn import vpn_validator
from neutron_vpnaas.services.vpn.plugin import VPNDriverPlugin
from neutron_vpnaas.services.vpn.plugin import VPNPlugin
from neutron_vpnaas.services.vpn.service_drivers import base_ipsec

import oslo_messaging

LOG = nfp_logging.getLogger(__name__)
BASE_VPN_VERSION = '1.0'
AGENT_TYPE_VPN = 'NFP Vpn agent'
ACTIVE = 'ACTIVE'
DOWN = 'DOWN'
ERROR = 'ERROR'
TIMEOUT = 80


class VPNAgentHostingServiceNotFound(exceptions.NeutronException):
    message = _("VPN Agent hosting vpn service '%(vpnservice_id)s' not found")


class VPNAgentNotFound(exceptions.NeutronException):
    message = _("VPN Agent not found in agent_db")


class VPNPluginExt(VPNPlugin, agentschedulers_db.AgentSchedulerDbMixin):
    """
    Extends the base VPN Plugin class to inherit agentdb too.
    Required to get agent entry into the database.
    """

    def __init__(self):
        super(VPNPluginExt, self).__init__()


class NFPIPsecVPNDriverCallBack(base_ipsec.IPsecVpnDriverCallBack):
    """Callback for IPSecVpnDriver rpc."""

    target = oslo_messaging.Target(version=BASE_VPN_VERSION)

    def __init__(self, driver):
        super(NFPIPsecVPNDriverCallBack, self).__init__(driver)
        self.driver = driver

    def update_status(self, context, status):
        """Update status of vpnservices."""
        if 'ipsec_site_connections' not in status[0]:
            status[0]['ipsec_site_connections'] = {}
        plugin = self.driver.service_plugin
        plugin.update_status_by_agent(context, status)


class NFPIPsecVpnAgentApi(base_ipsec.IPsecVpnAgentApi):
    """API and handler for NFP IPSec plugin to agent RPC messaging."""
    target = oslo_messaging.Target(version=BASE_VPN_VERSION)

    def __init__(self, topic, default_version, driver):
        super(NFPIPsecVpnAgentApi, self).__init__(
            topic, default_version, driver)

    def _is_agent_hosting_vpnservice(self, agent):
        """
        In case we have agent running on each compute node.
        We have to write logic here to get
        the agent which is hosting this vpn service
        """
        host = agent['host']
        lhost = socket.gethostname()
        if host == lhost:
            return True
        return False

    def _get_agent_hosting_vpnservice(self, admin_context, vpnservice_id):
        filters = {'agent_type': [AGENT_TYPE_VPN]}
        agents = manager.NeutronManager.get_plugin().get_agents(
            admin_context, filters=filters)

        try:
            for agent in agents:
                if not agent['alive']:
                    continue
                res = self._is_agent_hosting_vpnservice(agent)
                if res is True:
                    return agent

            # valid vpn agent is not found, hostname comparison might be
            # failed. Return whichever agent is available.
            for agent in agents:
                if not agent['alive']:
                    continue
                return agent
        except Exception:
            raise VPNAgentNotFound()

        msg = ('No active vpn agent found. Configuration will fail.')
        LOG.error(msg)
        raise VPNAgentHostingServiceNotFound(vpnservice_id=vpnservice_id)

    def _agent_notification(self, context, method, vpnservice_id,
                            version=None, **kwargs):
        admin_context = context.is_admin and context or context.elevated()

        if not version:
            version = self.target.version
        vpn_agent = self._get_agent_hosting_vpnservice(
            admin_context, vpnservice_id)

        msg = (('Notify agent at %(topic)s.%(host)s the message '
                '%(method)s %(args)s')
               % {'topic': self.topic,
                  'host': vpn_agent['host'],
                  'method': method, 'args': kwargs})
        LOG.debug(msg)

        cctxt = self.client.prepare(server=vpn_agent['host'],
                                    version=version)
        cctxt.cast(context, method, **kwargs)

    def vpnservice_updated(self, context, vpnservice_id, **kwargs):
        """
        Make rpc to agent for 'vpnservice_updated'
        """
        try:
            self._agent_notification(
                context, 'vpnservice_updated',
                vpnservice_id, **kwargs)
        except Exception:
            msg = ('Notifying agent failed')
            LOG.error(msg)


class VPNValidator(vpn_validator.VpnReferenceValidator):
    """This class overrides the vpnservice validator method"""
    def __init__(self):
        super(VPNValidator, self).__init__()

    def validate_vpnservice(self, context, vpns):
        pass


class NFPIPsecVPNDriver(base_ipsec.BaseIPsecVPNDriver):
    """VPN Service Driver class for IPsec."""

    def __init__(self, service_plugin):
        super(NFPIPsecVPNDriver, self).__init__(
            service_plugin)
        self.validator = VPNValidator()

    def create_rpc_conn(self):
        self.endpoints = [
            NFPIPsecVPNDriverCallBack(self),
            agents_db.AgentExtRpcCallback(VPNPluginExt())]

        self.conn = n_rpc.create_connection(new=True)
        self.conn.create_consumer(
            topics.VPN_NFP_PLUGIN_TOPIC, self.endpoints, fanout=False)
        self.conn.consume_in_threads()
        self.agent_rpc = NFPIPsecVpnAgentApi(
            topics.VPN_NFP_CONFIGAGENT_TOPIC, BASE_VPN_VERSION, self)

    def _get_service_vendor(self, context, vpnservice_id):
        vpnservice = self.service_plugin.get_vpnservice(
                context, vpnservice_id)
        desc = vpnservice['description']
        # if the call is through GBP workflow,
        # fetch the service profile from description
        # else, use 'VYOS' as the service profile
        if 'service_vendor=' in desc:
            tokens = desc.split(';')
            service_vendor = tokens[5].split('=')[1]
        else:
            service_vendor = 'VYOS'
        return service_vendor

    def create_ipsec_site_connection(self, context, ipsec_site_connection):
        service_vendor = self._get_service_vendor(
                                    context,
                                    ipsec_site_connection['vpnservice_id'])

        starttime = 0
        while starttime < TIMEOUT:
            vpnservice = self.service_plugin.get_vpnservice(
                                        context,
                                        ipsec_site_connection['vpnservice_id'])
            #(Revisit):Due to device driver issue neutron is making vpnservice
            #          state in Down state, At this point of time,
            #           Allowing ipsec site connection to gets created though
            #          vpnservice is in down state.
            if vpnservice['status'] in [ACTIVE, DOWN]:
                self.agent_rpc.vpnservice_updated(
                    context,
                    ipsec_site_connection['vpnservice_id'],
                    rsrc_type='ipsec_site_connection',
                    svc_type=self.service_type,
                    rsrc_id=ipsec_site_connection['id'],
                    resource=ipsec_site_connection,
                    reason='create', service_vendor=service_vendor)
                break
            elif vpnservice['status'] == ERROR:
                msg = ('updating ipsec_site_connection with id %s to'
                       'ERROR state' % (ipsec_site_connection['id']))
                LOG.error(msg)
                VPNPluginExt().update_ipsec_site_conn_status(
                                            context,
                                            ipsec_site_connection['id'],
                                            ERROR)
                break
            time.sleep(5)
            starttime += 5
        else:
            msg = ('updating ipsec_site_connection with id %s to'
                   'ERROR state' % (ipsec_site_connection['id']))
            LOG.error(msg)
            VPNPluginExt().update_ipsec_site_conn_status(
                                                context,
                                                ipsec_site_connection['id'],
                                                ERROR)

    def _move_ipsec_conn_state_to_error(self, context, ipsec_site_connection):
        vpnsvc_status = [{
            'id': ipsec_site_connection['vpnservice_id'],
            'status':ERROR,
            'updated_pending_status':False,
            'ipsec_site_connections':{
                ipsec_site_connection['id']: {
                    'status': ERROR,
                    'updated_pending_status': True}}}]
        driver = VPNDriverPlugin()._get_driver_for_ipsec_site_connection(
                                                    context,
                                                    ipsec_site_connection)
        NFPIPsecVPNDriverCallBack(driver).update_status(context,
                                                        vpnsvc_status)

    def delete_ipsec_site_connection(self, context, ipsec_site_connection):
        service_vendor = self._get_service_vendor(
                                    context,
                                    ipsec_site_connection['vpnservice_id'])

        self.agent_rpc.vpnservice_updated(
            context,
            ipsec_site_connection['vpnservice_id'],
            rsrc_type='ipsec_site_connection',
            svc_type=self.service_type,
            rsrc_id=ipsec_site_connection['id'],
            resource=ipsec_site_connection,
            reason='delete', service_vendor=service_vendor)

    def create_vpnservice(self, context, vpnservice):
        service_vendor = self._get_service_vendor(context,
                                                  vpnservice['id'])

        self.agent_rpc.vpnservice_updated(
            context,
            vpnservice['id'],
            rsrc_type='vpn_service',
            svc_type=self.service_type,
            rsrc_id=vpnservice['id'],
            resource=vpnservice,
            reason='create', service_vendor=service_vendor)

    def delete_vpnservice(self, context, vpnservice):
        pass

    def update_vpnservice(self, context, old_vpnservice, new_vpnservice):
        pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from keystoneclient.v2_0 import client as identity_client
import mock
from oslo_config import cfg
from oslo_utils import uuidutils
import unittest

from gbpclient.v2_0 import client as gbp_client
from gbpservice.neutron.tests.unit.nfp.orchestrator import mock_dicts
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.orchestrator.config_drivers import (
    heat_client as heat_client)
from gbpservice.nfp.orchestrator.config_drivers import heat_driver
from neutronclient.v2_0 import client as neutron_client


class MockStackObject(object):

    def __init__(self, status):
        self.stack_status = status


class MockHeatClientFunctionsDeleteNotFound(object):

    def delete(self, stack_id):
        raise heat_client.exc.HTTPNotFound()

    def create(self, **fields):
        return {'stack': {'id': uuidutils.generate_uuid()}}

    def get(self, stack_id):
        return MockStackObject('DELETE_COMPLETE')


class MockHeatClientFunctions(object):

    def delete(self, stack_id):
        pass

    def create(self, **fields):
        return {'stack': {'id': uuidutils.generate_uuid()}}

    def get(self, stack_id):
        return MockStackObject('DELETE_COMPLETE')

    def update(self, *args, **fields):
        return {'stack': {'id': uuidutils.generate_uuid()}}


class MockHeatClient(object):

    def __init__(self, api_version, endpoint, **kwargs):
        self.stacks = MockHeatClientFunctions()

cfg.CONF.import_group('nfp_keystone_authtoken',
                      'gbpservice.nfp.orchestrator.modules.__init__')
IS_SERVICE_ADMIN_OWNED = True
SVC_MGMT_PTG_NAME = 'svc_management_ptg'
RESOURCE_OWNER_TENANT_ID = '8ae6701128994ab281dde6b92207bb19'


class TestHeatDriver(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(TestHeatDriver, self).__init__(*args, **kwargs)
        with mock.patch.object(identity_client, "Client"):
            self.heat_driver_obj = heat_driver.HeatDriver(cfg.CONF)
        self.mock_dict = mock_dicts.DummyDictionaries()

    def setUp(self):
        cfg.CONF.set_override('is_service_admin_owned',
                              IS_SERVICE_ADMIN_OWNED,
                              group='heat_driver')
        cfg.CONF.set_override('svc_management_ptg_name',
                              SVC_MGMT_PTG_NAME,
                              group='heat_driver')
        cfg.CONF.set_override('admin_user',
                              'neutron',
                              group='nfp_keystone_authtoken')
        cfg.CONF.set_override('admin_password',
                              'admin_pass',
                              group='nfp_keystone_authtoken')
        cfg.CONF.set_override('admin_tenant_name',
                              'admin',
                              group='nfp_keystone_authtoken')
        # cfg.CONF.set_override('resource_owner_tenant_id',
        #                      RESOURCE_OWNER_TENANT_ID,
        #                      group='heat_driver')
        mock.patch(heat_client.__name__ + ".HeatClient",
                   new=MockHeatClient).start()

    @mock.patch.object(identity_client, "Client")
    def test_keystone(self, mock_obj):
        keystone_client = mock_obj.return_value
        keystone_client.auth_token = 'abcd123'
        password = 'neutron_pass'
        tenant_name = 'services'
        username = 'neutron'
        expected_token = 'abcd123'
        token = self.heat_driver_obj.keystone(
            username, password, tenant_name, tenant_id=None)
        self.assertEqual(token, expected_token)

    @mock.patch.object(identity_client, "Client")
    def test_get_heat_client(self, mock_obj):
        keystone_client = mock_obj.return_value
        keystone_client.auth_token = True
        self.heat_driver_obj._assign_admin_user_to_project = mock.Mock(
            return_value=None)
        nfp_logging.get_logging_context = mock.Mock(
            return_value={'auth_token': '7fd6701128994ab281ccb6b92207bb15'})
        tenant_id = '8ae6701128994ab281dde6b92207bb19'
        heat_client_obj = self.heat_driver_obj._get_heat_client(
            tenant_id)
        self.assertIsNotNone(heat_client_obj)

    @mock.patch.object(identity_client, "Client")
    def test_resource_owner_tenant_id(self, mock_obj):
        keystone_client = mock_obj.return_value
        keystone_client.auth_token = True
        keystone_client.tenants.find().id = '8ae6701128994ab281dde6b92207bb19'
        expected_resource_owner_tenant_id = '8ae6701128994ab281dde6b92207bb19'
        resource_owner_tenant_id = (
            self.heat_driver_obj._resource_owner_tenant_id())
        self.assertEqual(resource_owner_tenant_id,
                         expected_resource_owner_tenant_id)

    def mock_objects(self):
        with mock.patch.object(identity_client, "Client"):
            self.heat_driver_obj = heat_driver.HeatDriver(cfg.CONF)
            self.heat_driver_obj.keystoneclient.get_scoped_keystone_token = (
                mock.MagicMock(return_value='token'))
            self.heat_driver_obj.keystoneclient.get_tenant_id = mock.MagicMock(
                return_value='8ae6701128994ab281dde6b92207bb19')
            self.heat_driver_obj.neutron_client.get_port = mock.MagicMock(
                return_value=self.mock_dict.port_info)
            self.heat_driver_obj.neutron_client.get_floating_ips = (
                mock.MagicMock(return_value=self.mock_dict.fip))
            self.heat_driver_obj.neutron_client.get_subnets = mock.MagicMock(
                return_value=self.mock_dict.subnets_info['subnets'])
            self.heat_driver_obj.neutron_client.get_subnet = mock.MagicMock(
                return_value=self.mock_dict.subnet_info)
            self.heat_driver_obj.gbp_client.get_external_policies = (
                mock.MagicMock(
                    return_value=self.mock_dict.external_policies[
                        'external_policies']))
            self.heat_driver_obj.gbp_client.get_network_service_policies = (
                mock.MagicMock(return_value=self.mock_dict.services_nsp))
            self.heat_driver_obj.gbp_client.get_l3_policies = mock.MagicMock(
                return_value=self.mock_dict.l3_policies['l3_policies'])
            self.heat_driver_obj.gbp_client.get_policy_targets = (
                mock.MagicMock(
                    return_value=self.mock_dict.policy_targets[
                        'policy_targets']))
            self.heat_driver_obj.gbp_client.get_policy_target_groups = (
                mock.MagicMock(
                    return_value=self.mock_dict.policy_target_groups[
                        'policy_target_groups']))
            self.heat_driver_obj.gbp_client.get_policy_rule_sets = (
                mock.MagicMock(
                    return_value=self.mock_dict.policy_rule_sets[
                        'policy_rule_sets']))
            self.heat_driver_obj.gbp_client.get_policy_rules = mock.MagicMock(
                return_value=self.mock_dict.policy_rules['policy_rules'])
            self.heat_driver_obj.gbp_client.get_policy_actions = (
                mock.MagicMock(
                    return_value=self.mock_dict.policy_actions[
                        'policy_actions']))
            self.heat_driver_obj.gbp_client.get_l3_policy = mock.MagicMock(
                return_value=self.mock_dict.l3p)
            self.heat_driver_obj.gbp_client.get_l2_policy = mock.MagicMock(
                return_value=self.mock_dict.l2p)
            self.heat_driver_obj.gbp_client.update_policy_target_group = (
                mock.MagicMock(return_value={}))
            self.heat_driver_obj.gbp_client.create_policy_target_group = (
                mock.MagicMock(return_value={}))
            self.heat_driver_obj.gbp_client.create_policy_target = (
                mock.MagicMock(return_value=self.mock_dict.policy_target))
            self.heat_driver_obj.gbp_client.create_network_service_policy = (
                mock.MagicMock(return_value={}))

    def test_get_resource_owner_context(self):
        self.mock_objects()
        expected_resource_owner_context = (
            'token', '8ae6701128994ab281dde6b92207bb19')
        resource_owner_context = (
            self.heat_driver_obj._get_resource_owner_context())
        self.assertEqual(resource_owner_context,
                         expected_resource_owner_context)

    @mock.patch.object(identity_client, "Client")
    def test_get_tenant_context(self, mock_obj):
        keystone_client = mock_obj.return_value
        keystone_client.auth_token = True
        tenant_id = '8ae6701128994ab281dde6b92207bb19'
        expected_tenant_context = (True, tenant_id)
        tenant_context = self.heat_driver_obj._get_tenant_context(tenant_id)
        self.assertEqual(tenant_context, expected_tenant_context)

    def test_is_service_target(self):
        policy_target = {'name': 'service_target_provider_0132c_00b93'}
        retval = self.heat_driver_obj._is_service_target(policy_target)
        self.assertTrue(retval)
        policy_target = {'name': 'mem1_gbpui'}
        expected_result = False
        result = self.heat_driver_obj._is_service_target(policy_target)
        self.assertEqual(result, expected_result)

    @mock.patch.object(neutron_client.Client, "show_port")
    @mock.patch.object(gbp_client.Client, "list_policy_targets")
    def test_get_member_ips(self, list_pt_mock_obj, show_port_mock_obj):
        list_pt_mock_obj.return_value = self.mock_dict.policy_targets
        show_port_mock_obj.return_value = self.mock_dict.port_info
        auth_token = "81273djs138"
        expected_member_ips = ['42.0.0.13']
        member_ips = self.heat_driver_obj._get_member_ips(
            auth_token, self.mock_dict.provider_ptg)
        self.assertEqual(member_ips, expected_member_ips)

    def test_generate_lb_member_template(self):
        is_template_aws_version = False
        member_ip = '11.0.0.4'
        pool_res_name = 'HaproxyPool'
        stack_template = self.mock_dict.DEFAULT_LB_CONFIG
        expected_member_template = {
            'type': 'OS::Neutron::PoolMember',
            'properties': {
                'protocol_port': 101, 'admin_state_up': True,
                'pool_id': {'get_resource': 'HaproxyPool'},
                'weight': 1, 'address': '11.0.0.4'
            }
        }
        member_template = self.heat_driver_obj._generate_lb_member_template(
            is_template_aws_version,
            pool_res_name, member_ip, stack_template)
        self.assertEqual(member_template, expected_member_template)

    def test_modify_fw_resources_name(self):
        is_template_aws_version = False
        stack_template = copy.deepcopy(self.mock_dict.DEFAULT_FW_CONFIG)
        expected_fw_resources_name = 'serviceVM_infra_FW-fw_redirect'
        self.heat_driver_obj._modify_fw_resources_name(
            stack_template, self.mock_dict.provider_ptg,
            is_template_aws_version)
        modified_fw_resources_name = (
            stack_template['resources']['sc_firewall']['properties']['name'])
        self.assertEqual(modified_fw_resources_name,
                         expected_fw_resources_name)

    def test_get_heat_resource_key(self):
        is_template_aws_version = False
        resource_name = 'OS::Neutron::Pool'
        template_resource_dict = self.mock_dict.DEFAULT_LB_CONFIG['resources']
        expected_heat_resource_key = 'LoadBalancerPool'
        heat_resource_key = self.heat_driver_obj._get_heat_resource_key(
            template_resource_dict, is_template_aws_version, resource_name)
        self.assertEqual(heat_resource_key, expected_heat_resource_key)

    def test_get_all_heat_resource_keys(self):
        is_template_aws_version = False
        resource_name = 'OS::Neutron::Pool'
        template_resource_dict = self.mock_dict.DEFAULT_LB_CONFIG['resources']
        expected_heat_resource_keys = ['LoadBalancerPool']
        all_heat_resource_keys = (
            self.heat_driver_obj._get_all_heat_resource_keys(
                template_resource_dict, is_template_aws_version,
                resource_name))
        self.assertEqual(all_heat_resource_keys, expected_heat_resource_keys)

    @mock.patch.object(neutron_client.Client, "show_port")
    @mock.patch.object(gbp_client.Client, "list_policy_targets")
    def test_generate_pool_members(self, list_pt_mock_obj, show_port_mock_obj):
        list_pt_mock_obj.return_value = self.mock_dict.policy_targets
        show_port_mock_obj.return_value = self.mock_dict.port_info
        is_template_aws_version = False
        stack_template = self.mock_dict.DEFAULT_LB_CONFIG
        auth_token = "81273djs138"
        config_param_values = {}
        expected_pool_members = self.mock_dict.pool_members
        self.heat_driver_obj._generate_pool_members(
            auth_token,
            stack_template,
            config_param_values,
            self.mock_dict.provider_ptg,
            is_template_aws_version)
        generated_pool_members = stack_template['resources']['mem-42.0.0.13']
        self.assertEqual(generated_pool_members, expected_pool_members)

    def test_append_firewall_rule(self):
        stack_template = copy.deepcopy(self.mock_dict.DEFAULT_FW_CONFIG)
        provider_cidr = '192.169.0.0/29'
        consumer_cidr = '11.0.2.0/24'
        consumer_id = '2b86019a-45f7-4441-8e2c-1fbded4432c1'
        self.heat_driver_obj._append_firewall_rule(
            stack_template,
            provider_cidr, consumer_cidr,
            self.mock_dict.fw_template_properties,
            consumer_id)
        self.assertEqual(stack_template['resources']['sc_firewall_policy'],
                         self.mock_dict.appended_sc_firewall_policy)

    @mock.patch.object(heat_client.HeatClient, 'delete')
    @mock.patch.object(heat_client.HeatClient, 'get')
    @mock.patch.object(identity_client, "Client")
    def test_delete_config(self, mock_obj, heat_get_mock_obj,
                           heat_delete_mock_obj):
        heat_get_mock_obj.return_value = MockStackObject('DELETE_COMPLETE')
        self.heat_driver_obj._assign_admin_user_to_project = mock.Mock(
            return_value=None)
        nfp_logging.get_logging_context = mock.Mock(
            return_value={'auth_token': '7fd6701128994ab281ccb6b92207bb15'})

        instance = mock_obj.return_value
        instance.auth_token = True
        stack_id = '70754fdd-0325-4856-8a39-f171b65617d6'
        self.heat_driver_obj.delete_config(stack_id, '1627')
        heat_delete_mock_obj.assert_called_once_with(stack_id)

    @mock.patch.object(heat_client.HeatClient, 'get')
    @mock.patch.object(identity_client, "Client")
    def test_is_config_complete(self, mock_obj, heat_get_mock_obj):
        stack_id = '70754fdd-0325-4856-8a39-f171b65617d6'
        tenant_id = '8ae6701128994ab281dde6b92207bb19'
        self.heat_driver_obj._assign_admin_user_to_project = mock.Mock(
            return_value=None)
        nfp_logging.get_logging_context = mock.Mock(
            return_value={'auth_token': '7fd6701128994ab281ccb6b92207bb15'})
        self.heat_driver_obj.loadbalancer_post_stack_create = mock.Mock(
            return_value=None)
        heat_get_mock_obj.return_value = MockStackObject(
            'CREATE_COMPLETE')
        instance = mock_obj.return_value
        instance.auth_token = True
        expected_status = 'COMPLETED'
        status = self.heat_driver_obj.is_config_complete(
            stack_id, tenant_id, self.mock_dict.network_function_details)
        self.assertEqual(status, expected_status)

    @mock.patch.object(heat_client.HeatClient, 'get')
    @mock.patch.object(identity_client, "Client")
    def test_is_config_delete_complete(self, identity_mock_obj,
                                       heat_get_mock_obj):
        stack_id = '70754fdd-0325-4856-8a39-f171b65617d6'
        tenant_id = '8ae6701128994ab281dde6b92207bb19'
        self.heat_driver_obj._assign_admin_user_to_project = mock.Mock(
            return_value=None)
        nfp_logging.get_logging_context = mock.Mock(
            return_value={'auth_token': '7fd6701128994ab281ccb6b92207bb15'})
        heat_get_mock_obj.return_value = MockStackObject(
            'DELETE_COMPLETE')
        identity_mock_obj.return_value.auth_token = "1234"
        identity_mock_obj.return_value.tenants.find(
        ).id = "8ae6701128994ab281dde6b92207bb19"
        expected_status = 'COMPLETED'
        status = self.heat_driver_obj.is_config_delete_complete(stack_id,
                                                                tenant_id)
        self.assertEqual(status, expected_status)

    def test_get_site_conn_keys(self):
        is_template_aws_version = False
        resource_name = 'OS::Neutron::IPsecSiteConnection'
        template_resource_dict = (
            self.mock_dict.DEFAULT_VPN_CONFIG['resources'])
        expected_site_conn_keys = ['site_to_site_connection1']
        site_conn_keys = self.heat_driver_obj._get_site_conn_keys(
            template_resource_dict,
            is_template_aws_version, resource_name)
        self.assertEqual(site_conn_keys, expected_site_conn_keys)

    @mock.patch.object(neutron_client.Client, "show_subnet")
    @mock.patch.object(gbp_client.Client, "list_policy_target_groups")
    def test_get_management_gw_ip(self, list_ptg_mock_obj,
                                  show_subnet_mock_obj):
        list_ptg_mock_obj.return_value = self.mock_dict.policy_target_groups
        show_subnet_mock_obj.return_value = self.mock_dict.subnet_info
        auth_token = 'jkijqe18381'
        expected_management_gw_ip = '42.0.0.1'
        management_gw_ip = self.heat_driver_obj._get_management_gw_ip(
            auth_token)
        self.assertEqual(management_gw_ip, expected_management_gw_ip)

    @mock.patch.object(gbp_client.Client, "list_policy_actions")
    @mock.patch.object(gbp_client.Client, "list_policy_rules")
    @mock.patch.object(gbp_client.Client, "list_policy_rule_sets")
    def test_get_consumers_for_chain(self, list_policy_rule_sets_mock_obj,
                                     list_policy_rules_mock_obj,
                                     list_policy_actions_mock_obj):
        list_policy_rule_sets_mock_obj.return_value = (
            self.mock_dict.policy_rule_sets)
        list_policy_rules_mock_obj.return_value = self.mock_dict.policy_rules
        list_policy_actions_mock_obj.return_value = (
            self.mock_dict.policy_actions)
        auth_token = 'jkijqe18381'
        expected_consumers_for_chain = (
            (['af6a8a58-1e25-49c4-97a3-d5f50b3aa04b'], None))
        consumers_for_chain = self.heat_driver_obj._get_consumers_for_chain(
            auth_token,
            self.mock_dict.provider_ptg)
        self.assertEqual(consumers_for_chain, expected_consumers_for_chain)

    def test_update_firewall_template(self):
        self.mock_objects()
        stack_template = copy.deepcopy(self.mock_dict.DEFAULT_FW_CONFIG)
        auth_token = 'adakjiq'
        stack_template = self.heat_driver_obj._update_firewall_template(
            auth_token,
            self.mock_dict.provider_ptg, stack_template)
        self.assertEqual(
            stack_template['resources']['sc_firewall_policy'],
            copy.deepcopy(self.mock_dict.updated_template_sc_firewall_policy))

    @mock.patch.object(gbp_client.Client, "create_policy_target")
    @mock.patch.object(gbp_client.Client, "update_policy_target")
    @mock.patch.object(neutron_client.Client, "list_subnets")
    @mock.patch.object(neutron_client.Client, "list_pools")
    @mock.patch.object(neutron_client.Client, "show_vip")
    def test_create_policy_target_for_vip(self, vip, pools, subnets,
            pt, pt_update):
        pt.return_value = {
            'policy_target': {
                'name': 'service_target_provider_0132c_00b93'
            }
        }
        subnets.return_value = self.mock_dict.subnets_info
        pools.return_value = {
            'pools': [{
                'vip_id': '1234'
            }]
        }
        vip.return_value = {
            'vip': {
                'port_id': '1234'
            }
        }
        auth_token = 'adsdsdd'
        provider_tenant_id = '8ae6701128994ab281dde6b92207bb19'
        provider = self.mock_dict.provider_ptg
        self.heat_driver_obj.gbp_client.get_policy_targets = (
                mock.MagicMock(
                    return_value=self.mock_dict.policy_targets[
                        'policy_targets']))
        self.heat_driver_obj.keystoneclient.get_admin_token = (
                mock.MagicMock(return_value='token'))
        self.heat_driver_obj._create_policy_target_for_vip(
            auth_token, provider_tenant_id, provider,
            'LOADBALANCER')
        pools.assert_called_once_with(
            subnet_id=[subnets.return_value['subnets'][0]['id']])

    @mock.patch.object(neutron_client.Client, "list_networks")
    def test_create_node_config_data_vpn(self, mock_list_networks):
        self.mock_objects()
        auth_token = 'asdasasd'
        tenant_id = '8ae6701128994ab281dde6b92207bb19'
        provider = self.mock_dict.provider_ptg
        consumer = self.mock_dict.provider_ptg
        provider_port = self.mock_dict.port_info['port']
        mgmt_ip = self.mock_dict.mgmt_ip
        stack_template, stack_params = (
            self.heat_driver_obj._create_node_config_data(
                auth_token, tenant_id,
                self.mock_dict.vpn_service_chain_node,
                self.mock_dict.service_chain_instance,
                provider, provider_port, consumer,
                self.mock_dict.consumer_port,
                self.mock_dict.network_function_details['network_function'],
                mgmt_ip,
                self.mock_dict.service_details))
        self.assertEqual(stack_template['resources']['VPNService'][
                            'properties']['name'], 'VPNService')
        self.assertEqual(stack_params['RouterId'],
                         self.mock_dict.l3p['routers'][0])

    def test_update_node_config(self):
        self.mock_objects()
        auth_token = 'asdasasd'
        tenant_id = '8ae6701128994ab281dde6b92207bb19'
        provider = self.mock_dict.provider_ptg
        provider_port = self.mock_dict.port_info['port']
        mgmt_ip = self.mock_dict.mgmt_ip
        stack_template, stack_params = (
            self.heat_driver_obj._update_node_config(
                auth_token, tenant_id,
                self.mock_dict.service_profile,
                self.mock_dict.fw_service_chain_node,
                self.mock_dict.service_chain_instance,
                provider, self.mock_dict.consumer_port,
                self.mock_dict.network_function_details['network_function'],
                provider_port, mgmt_ip))
        self.assertEqual(stack_template['resources']['sc_firewall_policy'],
                         self.mock_dict.updated_sc_firewall_policy)

    @mock.patch.object(heat_client.HeatClient, "delete")
    @mock.patch.object(heat_client.HeatClient, "update")
    @mock.patch.object(heat_client.HeatClient, "get")
    @mock.patch.object(heat_client.HeatClient, "create")
    def test_update(
            self, heat_create, heat_get, heat_update,
            heat_delete):
        self.mock_objects()
        heat_create.return_value = {'stack': {
            'id': uuidutils.generate_uuid()}}
        heat_delete.return_value = MockStackObject(
            'CREATE_COMPLETE')
        self.heat_driver_obj._assign_admin_user_to_project = mock.Mock(
            return_value=None)
        nfp_logging.get_logging_context = mock.Mock(
            return_value={'auth_token': '7fd6701128994ab281ccb6b92207bb15'})
        auth_token = 'dasddasda'
        resource_owner_tenant_id = '8ae6701128994ab281dde6b92207bb19'
        provider = self.mock_dict.provider_ptg
        provider_port = self.mock_dict.port_info['port']
        stack_id = '70754fdd-0325-4856-8a39-f171b65617d6'
        mgmt_ip = self.mock_dict.mgmt_ip

        service_details = {}
        service_details['service_profile'] = self.mock_dict.service_profile
        service_details['servicechain_node'] = (
            self.mock_dict.fw_service_chain_node)
        service_details['servicechain_instance'] = (
            self.mock_dict.service_chain_instance)
        service_details['policy_target_group'] = self.mock_dict.provider_ptg
        service_details['provider_ptg'] = self.mock_dict.provider_ptg
        service_details['consumer_ptg'] = self.mock_dict.consumer_ptg
        service_details['consumer_port'] = self.mock_dict.consumer_port
        service_details['provider_port'] = self.mock_dict.port_info['port']
        service_details['mgmt_ip'] = '11.3.4.5'
        service_details['config_policy_id'] = (
            '70754fdd-0325-4856-8a39-f171b65617d6')
        self.heat_driver_obj.get_service_details = mock.Mock(
            return_value=service_details)

        stack_id = self.heat_driver_obj._update(
            auth_token, resource_owner_tenant_id,
            self.mock_dict.service_profile,
            self.mock_dict.fw_service_chain_node,
            self.mock_dict.service_chain_instance, provider,
            self.mock_dict.consumer_port,
            self.mock_dict.network_function_details['network_function'],
            provider_port, stack_id, mgmt_ip=mgmt_ip,
            pt_added_or_removed=False)
        self.assertIsNotNone(stack_id)

    @mock.patch.object(heat_client.HeatClient, "delete")
    @mock.patch.object(heat_client.HeatClient, "update")
    @mock.patch.object(heat_client.HeatClient, "get")
    @mock.patch.object(heat_client.HeatClient, "create")
    def test_handle_consumer_ptg_operations(
            self, heat_create,
            heat_get, heat_update, heat_delete):

        self.mock_objects()
        heat_create.return_value = {'stack': {
            'id': uuidutils.generate_uuid()}}
        heat_delete.return_value = MockStackObject(
            'CREATE_COMPLETE')
        self.heat_driver_obj._assign_admin_user_to_project = mock.Mock(
            return_value=None)
        nfp_logging.get_logging_context = mock.Mock(
            return_value={'auth_token': '7fd6701128994ab281ccb6b92207bb15'})

        service_details = {}
        service_details['service_profile'] = self.mock_dict.service_profile
        service_details['servicechain_node'] = (
            self.mock_dict.fw_service_chain_node)
        service_details['servicechain_instance'] = (
            self.mock_dict.service_chain_instance)
        service_details['policy_target_group'] = self.mock_dict.provider_ptg
        service_details['provider_ptg'] = self.mock_dict.provider_ptg
        service_details['consumer_ptg'] = self.mock_dict.consumer_ptg
        service_details['consumer_port'] = self.mock_dict.consumer_port
        service_details['provider_port'] = self.mock_dict.port_info['port']
        service_details['mgmt_ip'] = '11.3.4.5'
        service_details['config_policy_id'] = (
            '70754fdd-0325-4856-8a39-f171b65617d6')
        self.heat_driver_obj.get_service_details = mock.Mock(
            return_value=service_details)

        policy_target_group = self.mock_dict.provider_ptg
        stack_id = self.heat_driver_obj.handle_consumer_ptg_operations(
            self.mock_dict.network_function_details,
            policy_target_group, "add")
        self.assertIsNotNone(stack_id)

    @mock.patch.object(heat_client.HeatClient, "delete")
    @mock.patch.object(heat_client.HeatClient, "update")
    @mock.patch.object(heat_client.HeatClient, "get")
    @mock.patch.object(heat_client.HeatClient, "create")
    def test_handle_policy_target_operations(
            self, heat_create,
            heat_get, heat_update, heat_delete):

        self.mock_objects()
        heat_create.return_value = {'stack': {
            'id': uuidutils.generate_uuid()}}
        heat_delete.return_value = MockStackObject(
            'CREATE_COMPLETE')
        self.heat_driver_obj._assign_admin_user_to_project = mock.Mock(
            return_value=None)
        nfp_logging.get_logging_context = mock.Mock(
            return_value={'auth_token': '7fd6701128994ab281ccb6b92207bb15'})

        service_details = {}
        service_details['service_profile'] = self.mock_dict.lb_service_profile
        service_details['servicechain_node'] = (
            self.mock_dict.lb_service_chain_node)
        service_details['servicechain_instance'] = (
            self.mock_dict.service_chain_instance)
        service_details['policy_target_group'] = self.mock_dict.provider_ptg
        service_details['provider_ptg'] = self.mock_dict.provider_ptg
        service_details['consumer_ptg'] = self.mock_dict.consumer_ptg
        service_details['consumer_port'] = self.mock_dict.consumer_port
        service_details['provider_port'] = self.mock_dict.port_info['port']
        service_details['mgmt_ip'] = '11.3.4.5'
        service_details['config_policy_id'] = (
            '70754fdd-0325-4856-8a39-f171b65617d6')
        self.heat_driver_obj.get_service_details = mock.Mock(
            return_value=service_details)
        policy_target = {
            'name': 'policy_target_0132c_00b93'
        }
        stack_id = self.heat_driver_obj.handle_policy_target_operations(
            self.mock_dict.network_function_details,
            policy_target, "add")
        self.assertIsNotNone(stack_id)

    @mock.patch.object(heat_client.HeatClient, "delete")
    @mock.patch.object(heat_client.HeatClient, "update")
    @mock.patch.object(heat_client.HeatClient, "get")
    @mock.patch.object(heat_client.HeatClient, "create")
    def test_apply_config(
            self, heat_create,
            heat_get, heat_update, heat_delete):

        self.mock_objects()
        heat_create.return_value = {'stack': {
            'id': uuidutils.generate_uuid()}}
        heat_delete.return_value = MockStackObject(
            'CREATE_COMPLETE')
        self.heat_driver_obj._assign_admin_user_to_project = mock.Mock(
            return_value=None)
        nfp_logging.get_logging_context = mock.Mock(
            return_value={'auth_token': '7fd6701128994ab281ccb6b92207bb15'})

        service_details = {}
        service_details['service_profile'] = self.mock_dict.service_profile
        service_details['servicechain_node'] = (
            self.mock_dict.fw_service_chain_node)
        service_details['servicechain_instance'] = (
            self.mock_dict.service_chain_instance)
        service_details['policy_target_group'] = self.mock_dict.provider_ptg
        service_details['provider_ptg'] = self.mock_dict.provider_ptg
        service_details['consumer_ptg'] = self.mock_dict.consumer_ptg
        service_details['consumer_port'] = self.mock_dict.consumer_port
        service_details['provider_port'] = self.mock_dict.port_info['port']
        service_details['mgmt_ip'] = '11.3.4.5'
        service_details['config_policy_id'] = (
            '70754fdd-0325-4856-8a39-f171b65617d6')
        self.heat_driver_obj.get_service_details = mock.Mock(
            return_value=service_details)
        stack_id = self.heat_driver_obj.apply_config(
            self.mock_dict.network_function_details)
        self.assertIsNotNone(stack_id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.db import model_base
import sqlalchemy as sa


class GroupProxyMapping(model_base.BASEV2):
    __tablename__ = 'gp_group_proxy_mappings'
    policy_target_group_id = sa.Column(
        sa.String(36), sa.ForeignKey('gp_policy_target_groups.id',
                                     ondelete="CASCADE"), primary_key=True)
    # A group can only be proxied by one single group
    proxied_group_id = sa.Column(sa.String(36),
                                 sa.ForeignKey('gp_policy_target_groups.id'))
    # A group can only proxy one single other group
    # REVISIT(ivar): Can a backref be put here instead?
    proxy_group_id = sa.Column(sa.String(36),
                               sa.ForeignKey('gp_policy_target_groups.id',
                                             ondelete="SET NULL"))
    proxy_type = sa.Column(sa.String(24))
    enforce_service_chains = sa.Column(sa.Boolean, default=True,
                                       nullable=False)


class ProxyGatewayMapping(model_base.BASEV2):
    __tablename__ = 'gp_proxy_gateway_mappings'
    policy_target_id = sa.Column(
        sa.String(36), sa.ForeignKey('gp_policy_targets.id',
                                     ondelete="CASCADE"), primary_key=True)
    proxy_gateway = sa.Column(sa.Boolean, nullable=False)
    group_default_gateway = sa.Column(sa.Boolean, nullable=False)


class ProxyIPPoolMapping(model_base.BASEV2):
    __tablename__ = 'gp_proxy_ip_pool_mapping'

    l3_policy_id = sa.Column(
        sa.String(36), sa.ForeignKey('gp_l3_policies.id', ondelete="CASCADE"),
        primary_key=True)
    proxy_ip_pool = sa.Column(sa.String(64), nullable=False)
    proxy_subnet_prefix_length = sa.Column(sa.Integer, nullable=False)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpservice.nfp.core.event import Event
from gbpservice.nfp.proxy_agent.notifications import pull


def events_init(sc, conf):
    """Register event with its handler."""
    evs = [
        Event(id='PULL_NOTIFICATIONS',
              handler=pull.PullNotification(sc, conf))]
    sc.register_events(evs)


def nfp_module_init(sc, conf):
    """Initialize module to register rpc & event handler"""
    events_init(sc, conf)


def nfp_module_post_init(sc, conf):
    """Post a event for pull notification after each periodic_task_interval"""
    ev = sc.new_event(id='PULL_NOTIFICATIONS',
                      key='PULL_NOTIFICATIONS')
    sc.post_event(ev)



# Copyright (c) 2015 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron._i18n import _LW
from neutron.common import constants as q_const
from neutron.common import exceptions as n_exc
from neutron.db import common_db_mixin
from neutron.db import extraroute_db
from neutron.db import l3_gwmode_db
from neutron.plugins.common import constants
from oslo_log import log as logging

LOG = logging.getLogger(__name__)

from gbpservice.neutron.services.l3_router import apic_driver


class ApicGBPL3ServicePlugin(common_db_mixin.CommonDbMixin,
        extraroute_db.ExtraRoute_db_mixin,
        l3_gwmode_db.L3_NAT_db_mixin):

    supported_extension_aliases = ["router", "ext-gw-mode", "extraroute"]

    def __init__(self):
        super(ApicGBPL3ServicePlugin, self).__init__()
        self._apic_driver = apic_driver.ApicGBPL3Driver(self)

    def _update_router_gw_info(self, context, router_id, info, router=None):
        super(ApicGBPL3ServicePlugin, self)._update_router_gw_info(
            context, router_id, info, router)
        if info and 'network_id' in info:
            filters = {'device_id': [router_id],
                       'device_owner': [q_const.DEVICE_OWNER_ROUTER_GW],
                       'network_id': [info['network_id']]}
            ports = self._core_plugin.get_ports(context.elevated(),
                                                filters=filters)
            self._core_plugin.update_port_status(
                context, ports[0]['id'], q_const.PORT_STATUS_ACTIVE)

    @staticmethod
    def get_plugin_type():
        return constants.L3_ROUTER_NAT

    @staticmethod
    def get_plugin_description():
        """Returns string description of the plugin."""
        return _("L3 Router Service Plugin for basic L3 using the APIC")

    def add_router_interface(self, context, router_id, interface_info):
        port = super(ApicGBPL3ServicePlugin, self).add_router_interface(
            context, router_id, interface_info)
        self._apic_driver.add_router_interface_postcommit(
            context, router_id, interface_info)
        return port

    def remove_router_interface(self, context, router_id, interface_info):
        self._apic_driver.remove_router_interface_precommit(
            context, router_id, interface_info)
        super(ApicGBPL3ServicePlugin, self).remove_router_interface(
            context, router_id, interface_info)

    # Floating IP API
    def create_floatingip(self, context, floatingip):
        result = None
        fip = floatingip['floatingip']
        if not fip.get('subnet_id'):
            self._apic_driver.create_floatingip_precommit(context, floatingip)
            nat_pool_list = getattr(context, 'nat_pool_list', [])
            for nat_pool in nat_pool_list:
                if not nat_pool:
                    continue
                fip['subnet_id'] = nat_pool['subnet_id']
                try:
                    result = super(ApicGBPL3ServicePlugin,
                                   self).create_floatingip(context, floatingip)
                except n_exc.IpAddressGenerationFailure as ex:
                    LOG.warning(_LW("Floating allocation failed: %s"),
                                ex.message)
                if result:
                    break
        if not result:
            result = super(ApicGBPL3ServicePlugin,
                           self).create_floatingip(context, floatingip)
        context.result = result
        self._apic_driver.create_floatingip_postcommit(context, floatingip)
        return result

    def update_floatingip(self, context, id, floatingip):
        self._apic_driver.update_floatingip_precommit(context, id, floatingip)
        result = super(ApicGBPL3ServicePlugin,
                       self).update_floatingip(context, id, floatingip)
        context.result = result
        self._apic_driver.update_floatingip_postcommit(context,
                                                       id, floatingip)
        return result

    def delete_floatingip(self, context, id):
        self._apic_driver.delete_floatingip_precommit(context, id)
        result = super(ApicGBPL3ServicePlugin,
                       self).delete_floatingip(context, id)
        self._apic_driver.delete_floatingip_postcommit(context, id)
        return result



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import threading


class LogContext(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        if self.data:
            return {
                'meta_id': self.data.get('meta_id', '-'),
                'nfi_id': self.data.get('nfi_id', '-'),
                'nfd_id': self.data.get('nfd_id', '-'),
                'path': self.data.get('path'),
                'auth_token': self.data.get('auth_token'),
                'namespace': self.data.get('namespace')
            }
        return self.data


class CoreContext(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        return {
            'log_context': LogContext(self.data.get('log_context')).purge(),
            'event_desc': self.data.get('event_desc')
        }


class NfpContext(object):

    def __init__(self, data):
        self.data = data

    def purge(self):
        return CoreContext(self.data).purge()


Context = threading.local()


def init_log_context():
    return {
        'meta_id': '-',
        'nfi_id': '-',
        'nfd_id': '-',
        'path': '-',
        'auth_token': None,
        'namespace': None
    }


def init(data=None):
    if not data:
        data = {}
    if 'log_context' not in data.keys():
        data['log_context'] = init_log_context()
    if 'event_desc' not in data.keys():
        data['event_desc'] = {}
    Context.context = NfpContext(data)
    context = getattr(Context, 'context')
    return context.data


def get():
    try:
        context = getattr(Context, 'context')
        return context.data
    except AttributeError:
        return init()


def purge():
    try:
        context = getattr(Context, 'context')
        return context.purge()
    except AttributeError:
        init()
        context = getattr(Context, 'context')
        return context.purge()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from apic_ml2.neutron.db import port_ha_ipaddress_binding as ha_ip_db

from neutron._i18n import _LE
from neutron._i18n import _LW
from neutron.common import rpc as n_rpc
from neutron.common import topics
from neutron.db import api as db_api
from neutron.plugins.ml2 import rpc as ml2_rpc
from opflexagent import rpc as o_rpc
from oslo_log import log

from gbpservice.neutron.services.grouppolicy.drivers.cisco.apic import (
    nova_client as nclient)

LOG = log.getLogger(__name__)


class AIMMappingRPCMixin(ha_ip_db.HAIPOwnerDbMixin):
    """RPC mixin for AIM mapping.

    Collection of all the RPC methods consumed by the AIM mapping.
    By defining the mixin requirements, we can potentially move the RPC
    handling between GBP and Neutron preserving the same code base. Such
    requirements might be easier to implement in some places (eg. won't
    require model extensions) compared to others, based on the visibility
    that each module has over the network abstraction.
    """

    def setup_opflex_rpc_listeners(self):
        self.notifier = o_rpc.AgentNotifierApi(topics.AGENT)
        LOG.debug("Set up Opflex RPC listeners.")
        self.opflex_endpoints = [
            o_rpc.GBPServerRpcCallback(self, self.notifier)]
        self.opflex_topic = o_rpc.TOPIC_OPFLEX
        self.opflex_conn = n_rpc.create_connection(new=True)
        self.opflex_conn.create_consumer(
            self.opflex_topic, self.opflex_endpoints, fanout=False)
        self.opflex_conn.consume_in_threads()

    @db_api.retry_db_errors
    def _retrieve_vrf_details(self, context, **kwargs):
        with context.session.begin(subtransactions=True):
            details = {'l3_policy_id': kwargs['vrf_id']}
            details['_cache'] = {}
            self._add_vrf_details(context, details['l3_policy_id'], details)
            details.pop('_cache', None)
        return details

    def _get_vrf_details(self, context, **kwargs):
        LOG.debug("APIC AIM handling _get_vrf_details for: %s", kwargs)
        try:
            return self._retrieve_vrf_details(context, **kwargs)
        except Exception as e:
            vrf = kwargs.get('vrf_id')
            LOG.error(_LE("An exception has occurred while retrieving vrf "
                          "gbp details for %s"), vrf)
            LOG.exception(e)
            return {'l3_policy_id': vrf}

    def get_vrf_details(self, context, **kwargs):
        return self._get_vrf_details(context, **kwargs)

    def request_vrf_details(self, context, **kwargs):
        return self._get_vrf_details(context, **kwargs)

    def get_gbp_details(self, context, **kwargs):
        LOG.debug("APIC AIM handling get_gbp_details for: %s", kwargs)
        try:
            return self._get_gbp_details(context, kwargs, kwargs.get('host'))
        except Exception as e:
            device = kwargs.get('device')
            LOG.error(_LE("An exception has occurred while retrieving device "
                          "gbp details for %s"), device)
            LOG.exception(e)
            return {'device': device}

    def request_endpoint_details(self, context, **kwargs):
        LOG.debug("APIC AIM handling get_endpoint_details for: %s", kwargs)
        try:
            request = kwargs.get('request')
            host = kwargs.get('host')
            result = {'device': request['device'],
                      'timestamp': request['timestamp'],
                      'request_id': request['request_id'],
                      'gbp_details': self._get_gbp_details(context, request,
                                                           host),
                      'neutron_details': ml2_rpc.RpcCallbacks(
                          None, None).get_device_details(context, **request)}
            return result
        except Exception as e:
            LOG.error(_LE("An exception has occurred while requesting device "
                          "gbp details for %s"), request.get('device'))
            LOG.exception(e)
            return None

    # Child class needs to support:
    # - self._send_port_update_notification(context, port)
    def ip_address_owner_update(self, context, **kwargs):
        if not kwargs.get('ip_owner_info'):
            return
        ports_to_update = self.update_ip_owner(kwargs['ip_owner_info'])
        for p in ports_to_update:
            LOG.debug("APIC ownership update for port %s", p)
            self._send_port_update_notification(context, p)

    # Things you need in order to run this Mixin:
    # - self._core_plugin: attribute that points to the Neutron core plugin;
    # - self._is_port_promiscuous(context, port): define whether or not
    # a port should be put in promiscuous mode;
    # - self._get_port_epg(context, port): returns the AIM EPG for the specific
    # port
    # for both Neutron and GBP.
    # - self._is_dhcp_optimized(context, port);
    # - self._is_metadata_optimized(context, port);
    # - self._set_dhcp_lease_time(details)
    @db_api.retry_db_errors
    def _get_gbp_details(self, context, request, host):
        with context.session.begin(subtransactions=True):
            device = request.get('device')

            core_plugin = self._core_plugin
            port_id = core_plugin._device_to_port_id(context, device)
            port_context = core_plugin.get_bound_port_context(context, port_id,
                                                              host)
            if not port_context:
                LOG.warning(_LW("Device %(device)s requested by agent "
                                "%(agent_id)s not found in database"),
                            {'device': port_id,
                             'agent_id': request.get('agent_id')})
                return {'device': request.get('device')}
            port = port_context.current

            # NOTE(ivar): removed the PROXY_PORT_PREFIX hack.
            # This was needed to support network services without hotplug.

            epg = self._get_port_epg(context, port)

            details = {'device': request.get('device'),
                       'enable_dhcp_optimization': self._is_dhcp_optimized(
                           context, port),
                       'enable_metadata_optimization': (
                           self._is_metadata_optimized(context, port)),
                       'port_id': port_id,
                       'mac_address': port['mac_address'],
                       'app_profile_name': epg.app_profile_name,
                       'tenant_id': port['tenant_id'],
                       'host': host,
                       # TODO(ivar): scope names, possibly through AIM or the
                       # name mapper
                       'ptg_tenant': epg.tenant_name,
                       'endpoint_group_name': epg.name,
                       'promiscuous_mode': self._is_port_promiscuous(context,
                                                                     port),
                       'extra_ips': [],
                       'floating_ip': [],
                       'ip_mapping': [],
                       # Put per mac-address extra info
                       'extra_details': {}}

            # Set VM name if needed.
            if port['device_owner'].startswith(
                    'compute:') and port['device_id']:
                vm = nclient.NovaClient().get_server(port['device_id'])
                details['vm-name'] = vm.name if vm else port['device_id']
            mtu = self._get_port_mtu(context, port)
            if mtu:
                details['interface_mtu'] = mtu

            # NOTE(ivar): having these methods cleanly separated actually makes
            # things less efficient by requiring lots of calls duplication.
            # we could alleviate this by passing down a cache that stores
            # commonly requested objects (like EPGs). 'details' itself could
            # be used for such caching.
            details['_cache'] = {}
            vrf = self._get_port_vrf(context, port, details)
            details['l3_policy_id'] = '%s %s' % (vrf.tenant_name, vrf.name)
            self._add_subnet_details(context, port, details)
            self._add_allowed_address_pairs_details(context, port, details)
            self._add_vrf_details(context, details['l3_policy_id'], details)
            self._add_nat_details(context, port, host, details)
            self._add_extra_details(context, port, details)
            self._add_segmentation_label_details(context, port, details)
            self._set_dhcp_lease_time(details)
            details.pop('_cache', None)

        LOG.debug("Details for port %s : %s", port['id'], details)
        return details

    def _get_owned_addresses(self, plugin_context, port_id):
        return set(self.ha_ip_handler.get_ha_ipaddresses_for_port(port_id))

    # Child class needs to support:
    # - self._get_subnet_details(context, port, details)
    def _add_subnet_details(self, context, port, details):
        # This method needs to define requirements for this Mixin's child
        # classes in order to fill the following result parameters:
        # - subnets;
        details['subnets'] = self._get_subnet_details(context, port, details)

    def _add_nat_details(self, context, port, host, details):
        # This method needs to define requirements for this Mixin's child
        # classes in order to fill the following result parameters:
        # - floating_ip;
        # - ip_mapping;
        # - host_snat_ips.
        (details['floating_ip'], details['ip_mapping'],
            details['host_snat_ips']) = self._get_nat_details(
                context, port, host, details)

    # Child class needs to support:
    # - self._get_aap_details(context, port, details)
    def _add_allowed_address_pairs_details(self, context, port, details):
        # This method needs to define requirements for this Mixin's child
        # classes in order to fill the following result parameters:
        # - allowed_address_pairs
        # This should take care of realizing whether a given address is
        # active in the specific port
        details['allowed_address_pairs'] = self._get_aap_details(context, port,
                                                                 details)

    # Child class needs to support:
    # - self._get_vrf_subnets(context, vrf_tenant_name, vrf_name, details):
    # Subnets managed by the specific VRF.
    def _add_vrf_details(self, context, vrf_id, details):
        # This method needs to define requirements for this Mixin's child
        # classes in order to fill the following result parameters:
        # - l3_policy_id;
        # - vrf_tenant;
        # - vrf_name;
        # - vrf_subnets.
        tenant_name, name = vrf_id.split(' ')
        details['vrf_tenant'] = tenant_name
        details['vrf_name'] = name
        details['vrf_subnets'] = self._get_vrf_subnets(context, tenant_name,
                                                       name, details)

    # Child class needs to support:
    # - self._get_segmentation_labels(context, port, details)
    def _add_segmentation_label_details(self, context, port, details):
        # This method needs to define requirements for this Mixin's child
        # classes in order to fill the following result parameters:
        # - segmentation_labels
        # apic_segmentation_label is a GBP driver extension configured
        # for the aim_mapping driver
        details['segmentation_labels'] = self._get_segmentation_labels(
            context, port, details)

    def _add_extra_details(self, context, port, details):
        # TODO(ivar): Extra details depend on HA and SC implementation
        # This method needs to define requirements for this Mixin's child
        # classes in order to fill per-mac address extra information.

        # What is an "End of the Chain" port for Neutron?
        pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from neutron.tests import base

from gbpservice.contrib.nfp.configurator.agents import nfp_service as ns
from gbpservice.contrib.nfp.configurator.lib import (
                                        nfp_service_constants as const)
from gbpservice.contrib.tests.unit.nfp.configurator.test_data import (
                                                nfp_service_test_data as fo)


class NfpServiceRpcManagerTestCase(base.BaseTestCase):
    """ Implement test cases for RPC manager methods of nfp service agent.

    """

    def __init__(self, *args, **kwargs):
        super(NfpServiceRpcManagerTestCase, self).__init__(
                                                        *args, **kwargs)
        self.fo = fo.FakeObjects()

    @mock.patch(__name__ + '.fo.FakeObjects.sc')
    @mock.patch(__name__ + '.fo.FakeObjects.conf')
    def _get_NfpServiceRpcManager_object(self, conf, sc):
        """ Retrieves RPC manager object of nfp service agent.

        :param sc: mocked service controller object of process model framework
        :param conf: mocked OSLO configuration file

        Returns: object of nfp service's RPC manager
        and service controller.
        """

        agent = ns.ConfigScriptRpcManager(sc, conf)
        return agent, sc

    def _test_event_creation(self, method):
        """ Tests event creation and enqueueing for create/delete
        operation of generic config agent's RPC manager.

        :param method: CREATE_NFP_SERVICE

        Returns: none
        """

        agent, sc = self._get_NfpServiceRpcManager_object()
        arg_dict = {'context': self.fo.context,
                    'resource_data': self.fo.kwargs}
        with mock.patch.object(
                    sc, 'new_event', return_value='foo') as mock_sc_event, (
             mock.patch.object(sc, 'post_event')) as mock_sc_rpc_event:
            actual_call = agent.run_nfp_service(self.fo.context,
                                                self.fo.kwargs)

            expected_cal = mock_sc_event.assert_called_with(
                                id=method, data=arg_dict, key=None)
            self.assertEqual(actual_call, expected_cal)
            mock_sc_rpc_event.assert_called_with('foo')

    def test_nfp_service_rpcmanager(self):
        """ Implements test case for run_nfp_service method
        of nfp service RPC manager.

        Returns: none
        """

        self._test_event_creation(const.CREATE_NFP_SERVICE_EVENT)


class NfpServiceEventHandlerTestCase(base.BaseTestCase):
    """ Implements test cases for event handler methods
    of nfp service agent.

    """

    def __init__(self, *args, **kwargs):
        super(NfpServiceEventHandlerTestCase, self).__init__(
                                                        *args, **kwargs)
        self.fo = fo.FakeObjects()
        self.context = {'notification_data': {},
                        'resource': 'interfaces'}

    @mock.patch(__name__ + '.fo.FakeObjects.rpcmgr')
    @mock.patch(__name__ + '.fo.FakeObjects.drivers')
    @mock.patch(__name__ + '.fo.FakeObjects.sc')
    def _get_nfp_service_event_handler_object(self, sc, drivers, rpcmgr):
        """ Retrieves event handler object of nfp service.

        :param sc: mocked service controller object of process model framework
        :param rpcmgr: object of configurator's RPC manager
        :param drivers: list of driver objects for nfp service agent

        Returns: object of nfp service's event handler
        """

        agent = ns.ConfigScriptEventHandler(sc, drivers, rpcmgr)
        return agent, sc

    def _test_handle_event(self, ev, result=const.UNHANDLED_RESULT):
        """ Test handle event method of nfp service agent.

        :param ev: event data which has to be actually sent by
        process framework.

        Returns: None
        """

        agent, sc = self._get_nfp_service_event_handler_object()
        driver = mock.Mock()

        with mock.patch.object(
                driver, 'run_heat', return_value=result) as mock_config_inte, (
             mock.patch.object(
                agent, '_get_driver', return_value=driver)):

            agent.handle_event(ev)

            mock_config_inte.assert_called_with(
                ev.data['context']['context'], ev.data['resource_data'])

    def test_create_nfp_service_handle_event_success(self):
        """ Implements positive test case for create_nfp_service method
        of nfp service event handler.

        Returns: none
        """

        ev = fo.FakeEventNfpService()
        ev.id = const.CREATE_NFP_SERVICE_EVENT
        self._test_handle_event(ev)

    def test_create_nfp_service_handle_event_failure(self):
        """ Implements negative test case for create_nfp_service method
        of nfp service event handler.

        Returns: none
        """

        ev = fo.FakeEventNfpService()
        ev.id = const.CREATE_NFP_SERVICE_EVENT
        self._test_handle_event(ev, const.FAILURE)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import time

from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.lib import nfp_exceptions

LOG = nfp_logging.getLogger(__name__)

sql_lock_support = True


class ContextManager(object):

    def __init__(self, session=None, suppress=tuple()):
        # suppress tuple holds the kind of exceptions
        # the we don't have re-raise
        self.session = session
        self.suppress = suppress

    def __enter__(self):
        pass

    def __exit__(self, Exptype, expvalue, traceback):

        if self.suppress and Exptype:
            if Exptype in self.suppress:
                return False
            for exception in self.suppress:
                if isinstance(Exptype, exception):
                    return False
        if not self.suppress and traceback:
            return True
        else:
            return False

    def retry(self, method, *args, **kwargs):
        tries = kwargs.pop('tries', 1)
        delay = 2
        backoff = 2
        while tries > 1:
            # Loop for 'tries-1' times and
            # the last time without any try-catch
            try:
                return method(*args, **kwargs)
            except Exception:
                msg = " %s retrying in %s seconds " % (self.__class__, delay)
                LOG.error(msg)

            time.sleep(delay)
            tries -= 1
            delay *= backoff
        return method(*args, **kwargs)


class NfpDbContextManager(ContextManager):

    def new(self, **kwargs):
        return NfpDbContextManager(**kwargs)

    def lock(self, session, method, *args, **kwargs):
        if not sql_lock_support:
            return method(session, *args, **kwargs)
        with session.begin(subtransactions=True):
            session.execute("SELECT GET_LOCK('nfp_db_lock', -1)")
            ret = method(session, *args, **kwargs)
            session.execute("SELECT RELEASE_LOCK('nfp_db_lock')")
            return ret

    def __enter__(self):
        super(NfpDbContextManager, self).__enter__()
        return self

    def __exit__(self, Exptype, expvalue, traceback):
        if super(NfpDbContextManager, self).__exit__(
                Exptype, expvalue, traceback):
            raise nfp_exceptions.DbException(Exptype, str(expvalue), traceback)

        # By default exit method returns False, if False is returned
        # the with block re-raises the exception. To suppress that
        # True should be returned explicitly

        return True


class NfpNovaContextManager(ContextManager):

    def new(self, **kwargs):
        return NfpNovaContextManager(**kwargs)

    def __enter__(self):
        super(NfpNovaContextManager, self).__enter__()
        return self

    def __exit__(self, Exptype, expvalue, traceback):
        if super(NfpNovaContextManager, self).__exit__(
                Exptype, expvalue, traceback):
            raise nfp_exceptions.NovaException(
                Exptype, str(expvalue), traceback)

        # By default exit method returns False, if False is returned
        # the with block re-raises the exception. To suppress that
        # True should be returned explicitly

        return True


class NfpKeystoneContextManager(ContextManager):

    def new(self, **kwargs):
        return NfpKeystoneContextManager(**kwargs)

    def __enter__(self):
        super(NfpKeystoneContextManager, self).__enter__()
        return self

    def __exit__(self, Exptype, expvalue, traceback):
        if super(NfpKeystoneContextManager, self).__exit__(
                Exptype, expvalue, traceback):
            raise nfp_exceptions.KeystoneException(
                Exptype, str(expvalue), traceback)
        # By default exit method returns False, if False is returned
        # the with block re-raises the exception. To suppress that
        # True should be returned explicitly

        return True


class NfpNeutronContextManager(ContextManager):

    def new(self, **kwargs):
        return NfpNeutronContextManager(**kwargs)

    def __enter__(self):
        super(NfpNeutronContextManager, self).__enter__()
        return self

    def __exit__(self, Exptype, expvalue, traceback):
        if super(NfpNeutronContextManager, self).__exit__(
                Exptype, expvalue, traceback):
            raise nfp_exceptions.NeutronException(
                Exptype, str(expvalue), traceback)

        # By default exit method returns False, if False is returned
        # the with block re-raises the exception. To suppress that
        # True should be returned explicitly

        return True


class NfpHeatContextManager(ContextManager):

    def new(self, **kwargs):
        return NfpHeatContextManager(**kwargs)

    def __enter__(self):
        super(NfpHeatContextManager, self).__enter__()
        return self

    def __exit__(self, Exptype, expvalue, traceback):
        if super(NfpHeatContextManager, self).__exit__(
                Exptype, expvalue, traceback):
            raise nfp_exceptions.HeatException(
                Exptype, str(expvalue), traceback)

        # By default exit method returns False, if False is returned
        # the with block re-raises the exception. To suppress that
        # True should be returned explicitly

        return True


class NfpGBPContextManager(ContextManager):

    def new(self, **kwargs):
        return NfpGBPContextManager(**kwargs)

    def __enter__(self):
        super(NfpGBPContextManager, self).__enter__()
        return self

    def __exit__(self, Exptype, expvalue, traceback):
        if super(NfpGBPContextManager, self).__exit__(
                Exptype, expvalue, traceback):
            raise nfp_exceptions.GBPException(
                Exptype, str(expvalue), traceback)

        # By default exit method returns False, if False is returned
        # the with block re-raises the exception. To suppress that
        # True should be returned explicitly

        return True

# Create the respective instances once, so that no need
# to instantiate them again any where

DbContextManager = NfpDbContextManager()
NovaContextManager = NfpNovaContextManager()
KeystoneContextManager = NfpKeystoneContextManager()
NeutronContextManager = NfpNeutronContextManager()
HeatContextManager = NfpHeatContextManager()
GBPContextManager = NfpGBPContextManager()



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.extensions import address_scope


class Patched_address_scope(address_scope.Address_scope):
    def update_attributes_map(self, attributes):
        super(Patched_address_scope, self).update_attributes_map(
            attributes,
            extension_attrs_map=address_scope.RESOURCE_ATTRIBUTE_MAP)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import heatclient
import mock
from neutron.plugins.common import constants
from oslo_serialization import jsonutils
from oslo_utils import uuidutils
import webob

from gbpservice.neutron.services.servicechain.plugins.msc import config
from gbpservice.neutron.services.servicechain.plugins.msc.drivers import (
    simplechain_driver as simplechain_driver)
from gbpservice.neutron.tests.unit.services.servicechain import (
    test_servicechain_plugin as test_servicechain_plugin)


STACK_DELETE_RETRIES = 5
STACK_DELETE_RETRY_WAIT = 3


class MockStackObject(object):
    def __init__(self, status):
        self.stack_status = status


class MockHeatClientFunctions(object):
    def delete(self, stack_id):
        raise heatclient.exc.HTTPNotFound()

    def create(self, **fields):
        return {'stack': {'id': uuidutils.generate_uuid()}}

    def get(self, stack_id):
        return MockStackObject('DELETE_COMPLETE')

    def update(self, *args, **fields):
        return {'stack': {'id': uuidutils.generate_uuid()}}


class MockHeatClient(object):
    def __init__(self, api_version, endpoint, **kwargs):
        self.stacks = MockHeatClientFunctions()


class SimpleChainDriverTestCase(
        test_servicechain_plugin.ServiceChainPluginTestCase):

    def setUp(self):
        config.cfg.CONF.set_override('servicechain_drivers',
                                     ['simplechain_driver'],
                                     group='servicechain')
        config.cfg.CONF.set_override('stack_delete_retries',
                                     STACK_DELETE_RETRIES,
                                     group='simplechain')
        config.cfg.CONF.set_override('stack_delete_retry_wait',
                                     STACK_DELETE_RETRY_WAIT,
                                     group='simplechain')
        super(SimpleChainDriverTestCase, self).setUp()
        key_client = mock.patch(
            'gbpservice.neutron.services.servicechain.plugins.msc.drivers.'
            'simplechain_driver.HeatClient._get_auth_token').start()
        key_client.return_value = 'mysplendidtoken'


class TestServiceChainInstance(SimpleChainDriverTestCase):

    def test_invalid_service_type_rejected(self):
        res = self.create_service_profile(
                    service_type="test",
                    expected_res_status=webob.exc.HTTPBadRequest.code)
        self.assertEqual('InvalidServiceTypeForReferenceDriver',
                         res['NeutronError']['type'])

    def test_chain_node_create_success(self):
        res = self._create_profiled_servicechain_node(
                    service_type=constants.FIREWALL, config='{}',
                    expected_res_status=webob.exc.HTTPCreated.code)
        self.assertEqual('{}', res['servicechain_node']['config'])

    def test_in_use_node_config_update_rejected(self):
        node = self.create_servicechain_node(
            service_type=constants.FIREWALL, config='{}',
            expected_res_status=webob.exc.HTTPCreated.code)[
                                                'servicechain_node']
        self.assertEqual('{}', node['config'])
        spec = self.create_servicechain_spec(
            nodes=[node['id']],
            expected_res_status=webob.exc.HTTPCreated.code)[
                                                'servicechain_spec']
        with mock.patch.object(simplechain_driver.HeatClient,
                              'create') as stack_create:
            stack_create.return_value = {'stack': {
                                            'id': uuidutils.generate_uuid()}}
            self.create_servicechain_instance(servicechain_specs=[spec['id']])
            res = self.update_servicechain_node(
                    node['id'],
                    config='{"key": "value"}',
                    expected_res_status=webob.exc.HTTPBadRequest.code)
            self.assertEqual('NodeUpdateNotSupported',
                             res['NeutronError']['type'])

    def test_chain_spec_update(self):
        template1 = '{"key1":"value1"}'
        scn = self._create_profiled_servicechain_node(config=template1)
        scn1_name = scn['servicechain_node']['name']
        scn_id = scn['servicechain_node']['id']
        name = "scs1"
        template2 = '{"key2":"value2"}'
        scn2 = self._create_profiled_servicechain_node(config=template2)
        scn2_id = scn2['servicechain_node']['id']
        scn2_name = scn2['servicechain_node']['name']
        scs = self.create_servicechain_spec(name=name, nodes=[scn_id])
        sc_spec_id = scs['servicechain_spec']['id']

        stack1 = {'stack': {'id': uuidutils.generate_uuid()}}
        stack2 = {'stack': {'id': uuidutils.generate_uuid()}}
        stack3 = {'stack': {'id': uuidutils.generate_uuid()}}
        expected_create_calls = []
        expected_delete_calls = []
        with mock.patch.object(simplechain_driver.HeatClient,
                              'create') as stack_create:
            with mock.patch.object(simplechain_driver.HeatClient,
                                   'delete') as stack_delete:
                stack_create.return_value = stack1
                instance1_name = "sc_instance_1"
                sc_instance1 = self.create_servicechain_instance(
                                            name=instance1_name,
                                            servicechain_specs=[sc_spec_id])
                sci1_id = sc_instance1['servicechain_instance']['id']
                self.assertEqual([sc_spec_id],
                    sc_instance1['servicechain_instance'][
                        'servicechain_specs'])
                stack_name = ("stack_" + instance1_name + scn1_name +
                              sci1_id[:8])
                expected_create_calls.append(
                            mock.call(stack_name,
                                      jsonutils.loads(template1), {}))
                stack_create.return_value = stack2
                instance2_name = "sc_instance_2"
                sc_instance2 = self.create_servicechain_instance(
                                            name=instance2_name,
                                            servicechain_specs=[sc_spec_id])
                sci2_id = sc_instance2['servicechain_instance']['id']
                self.assertEqual(
                    [sc_spec_id],
                    sc_instance2['servicechain_instance'][
                        'servicechain_specs'])
                stack_name = ("stack_" + instance2_name + scn1_name +
                              sci2_id[:8])
                expected_create_calls.append(
                    mock.call(stack_name, jsonutils.loads(template1), {}))

                #Now perform an update of the spec
                new_spec = {'servicechain_spec': {'nodes': [scn2_id]}}
                stack_create.return_value = stack3
                req = self.new_update_request(
                            'servicechain_specs', new_spec, sc_spec_id)
                res = req.get_response(self.ext_api)
                self.assertEqual(webob.exc.HTTPOk.code, res.status_int)
                # The two existing stacks will be deleted and two new stacks
                # will be created
                expected_delete_calls.append(mock.call(stack1['stack']['id']))
                expected_delete_calls.append(mock.call(stack2['stack']['id']))
                stack_name = ("stack_" + instance1_name + scn2_name +
                              sci1_id[:8])
                expected_create_calls.append(
                    mock.call(stack_name, jsonutils.loads(template2), {}))
                stack_name = ("stack_" + instance2_name + scn2_name +
                              sci2_id[:8])
                expected_create_calls.append(
                    mock.call(stack_name, jsonutils.loads(template2), {}))
                self.assertEqual(expected_delete_calls,
                                 stack_delete.call_args_list)
                self.assertEqual(expected_create_calls,
                                 stack_create.call_args_list)

    def test_chain_instance_create(self):
        name = "scs1"
        scn = self._create_profiled_servicechain_node()
        scn_id = scn['servicechain_node']['id']
        scs = self.create_servicechain_spec(name=name, nodes=[scn_id])
        sc_spec_id = scs['servicechain_spec']['id']

        with mock.patch.object(simplechain_driver.HeatClient,
                               'create') as stack_create:
            stack_create.return_value = {'stack': {
                                        'id': uuidutils.generate_uuid()}}
            sc_instance = self.create_servicechain_instance(
                                        name="sc_instance_1",
                                        servicechain_specs=[sc_spec_id])
            expected_stack_name = (
                "stack_" + "sc_instance_1" +
                scn['servicechain_node']['name'] +
                sc_instance['servicechain_instance']['id'][:8])
            self.assertEqual(
                [sc_spec_id],
                sc_instance['servicechain_instance']['servicechain_specs'])
            stack_create.assert_called_once_with(
                                    expected_stack_name, mock.ANY, mock.ANY)

    def test_chain_instance_delete(self):
        name = "scs1"
        scn = self._create_profiled_servicechain_node()
        scn_id = scn['servicechain_node']['id']
        scs = self.create_servicechain_spec(name=name, nodes=[scn_id])
        sc_spec_id = scs['servicechain_spec']['id']

        with mock.patch.object(simplechain_driver.HeatClient,
                               'create') as stack_create:
            stack_create.return_value = {'stack': {
                                        'id': uuidutils.generate_uuid()}}
            sc_instance = self.create_servicechain_instance(
                                        name="sc_instance_1",
                                        servicechain_specs=[sc_spec_id])
            self.assertEqual([sc_spec_id],
                sc_instance['servicechain_instance']['servicechain_specs'])
            with mock.patch.object(simplechain_driver.HeatClient,
                                   'delete'):
                req = self.new_delete_request(
                                    'servicechain_instances',
                                    sc_instance['servicechain_instance']['id'])
                res = req.get_response(self.ext_api)
                self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

    def test_wait_stack_delete_for_instance_delete(self):
        name = "scs1"
        scn = self._create_profiled_servicechain_node()
        scn_id = scn['servicechain_node']['id']
        scs = self.create_servicechain_spec(name=name, nodes=[scn_id])
        sc_spec_id = scs['servicechain_spec']['id']

        with mock.patch.object(simplechain_driver.HeatClient,
                               'create') as stack_create:
            stack_create.return_value = {'stack': {
                                        'id': uuidutils.generate_uuid()}}
            sc_instance = self.create_servicechain_instance(
                                        name="sc_instance_1",
                                        servicechain_specs=[sc_spec_id])
            self.assertEqual([sc_spec_id],
                sc_instance['servicechain_instance']['servicechain_specs'])

            # Verify that as part of delete service chain instance we call
            # get method for heat stack 5 times before giving up if the state
            # does not become DELETE_COMPLETE
            with mock.patch.object(simplechain_driver.HeatClient,
                                   'delete') as stack_delete:
                with mock.patch.object(simplechain_driver.HeatClient,
                                       'get') as stack_get:
                    stack_get.return_value = MockStackObject('PENDING_DELETE')
                    req = self.new_delete_request(
                        'servicechain_instances',
                        sc_instance['servicechain_instance']['id'])
                    res = req.get_response(self.ext_api)
                    self.assertEqual(webob.exc.HTTPNoContent.code,
                                     res.status_int)
                    stack_delete.assert_called_once_with(mock.ANY)
                    self.assertEqual(STACK_DELETE_RETRIES,
                                     stack_get.call_count)

            # Create and delete another service chain instance and verify that
            # we call get method for heat stack only once if the stack state
            # is DELETE_COMPLETE
            sc_instance = self.create_servicechain_instance(
                                        name="sc_instance_1",
                                        servicechain_specs=[sc_spec_id])
            self.assertEqual(
                [sc_spec_id],
                sc_instance['servicechain_instance']['servicechain_specs'])
            with mock.patch.object(simplechain_driver.HeatClient,
                                   'delete') as stack_delete:
                with mock.patch.object(simplechain_driver.HeatClient,
                                       'get') as stack_get:
                    stack_get.return_value = MockStackObject(
                        'DELETE_COMPLETE')
                    req = self.new_delete_request(
                        'servicechain_instances',
                        sc_instance['servicechain_instance']['id'])
                    res = req.get_response(self.ext_api)
                    self.assertEqual(webob.exc.HTTPNoContent.code,
                                     res.status_int)
                    stack_delete.assert_called_once_with(mock.ANY)
                    self.assertEqual(1, stack_get.call_count)

    def test_stack_not_found_ignored(self):
        name = "scs1"
        scn = self._create_profiled_servicechain_node()
        scn_id = scn['servicechain_node']['id']
        scs = self.create_servicechain_spec(name=name, nodes=[scn_id])
        sc_spec_id = scs['servicechain_spec']['id']

        mock.patch(heatclient.__name__ + ".client.Client",
                   new=MockHeatClient).start()
        sc_instance = self.create_servicechain_instance(
                                    name="sc_instance_1",
                                    servicechain_specs=[sc_spec_id])
        req = self.new_delete_request(
                            'servicechain_instances',
                            sc_instance['servicechain_instance']['id'])
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" This class provides data that is needed for calling methods of
agent and driver.

"""


class Foo(object):

    key = 'key'
    serialize = 'serialize'
    binding_key = 'binding_key'


class Context(object):
    def to_dict(self):
        return {}


class FakeObjects(object):

    sc = 'sc'
    context = {'notification_data': {},
               'resource': 'context_resource'}
    context_pool = {'notification_data': {},
                    'resource': 'vip'}
    conf = 'conf'
    rpcmgr = 'rpcmgr'
    nqueue = 'nqueue'
    drivers = ['haproxy']
    vip_context = {'notification_data': {}, 'resource': 'context_resource'}
    context_test = {'notification_data': {}, 'resource': 'context_resource'}
    method = {'CREATE_VIP': 'create_network_function_config',
              'DELETE_VIP': 'delete_network_function_config',
              'UPDATE_VIP': 'update_network_function_config',
              'CREATE_POOL': 'create_network_function_config',
              'DELETE_POOL': 'delete_network_function_config',
              'UPDATE_POOL': 'update_network_function_config',
              'CREATE_MEMBER': 'create_network_function_config',
              'DELETE_MEMBER': 'delete_network_function_config',
              'UPDATE_MEMBER': 'update_network_function_config',
              'CREATE_POOL_HEALTH_MONITOR': 'create_network_function_config',
              'DELETE_POOL_HEALTH_MONITOR': 'delete_network_function_config',
              'UPDATE_POOL_HEALTH_MONITOR': 'update_network_function_config'}

    def _get_context_logical_device(self):
        context_logical_device = {
            'service_info': {
                'pools': self._get_pool_object(),
                'vips': self._get_vip_object(),
                'members': self._get_member_object(),
                'health_monitors': self._get_hm_object(),
                'ports': self._get_ports_object(),
                'subnets': self._get_subnets_object()}}
        return context_logical_device

    def lb_api_context(self):
            context = {
                'agent_info': {
                    'resource': 'firewall',
                    'service_vendor': 'vyos',
                    'context': {'requester': 'device_orch',
                                'logging_context': {}},
                    'resource_type': 'firewall'},
                'notification_data': {}, 'service_info': {},
                "resource_data": {
                        "forward_route": True,
                        "tenant_id": "ac33b4c2d80f485a86ea515c09c74949",
                        "nfs": [{
                            "role": "master",
                            "svc_mgmt_fixed_ip": "11.0.0.37",
                            "networks": [
                                {"cidr": "11.0.1.0/24",
                                 "gw_ip": "",
                                 "type": "provider",
                                 "ports": [{
                                    "mac": "fa:16:3e:d9:4c:33",
                                    "fixed_ip": "11.0.1.1",
                                    "floating_ip": ""}]},
                                {"cidr": "192.168.0.0/28",
                                 "gw_ip": "192.168.0.1 ",
                                 "type": "stitching",
                                 "ports": [{
                                     "mac": "fa:16:3e:da:ca:4d",
                                     "fixed_ip": "192.168.0.3",
                                     "floating_ip": ""}]}
                                            ]}]}}
            return context

    def get_request_data_for_vip(self):
        """Returns request data needed for create_vip method.

        Returns: request_data

        """

        request_data = {
            "info": {'context': {"logging_context": {}},
                     'service_type': "loadbalancer",
                     'service_vendor': "haproxy",
                     },
            "config": [{
                "resource": "vip",
                "resource_data": {
                    "neutron_context": self.context,
                    "vip": self._get_vip_object()[0]
                }}]}
        return request_data

    def get_request_data_for_vip_update(self):
        """Returns request data needed for update_vip method.

        Returns: request_data

        """

        request_data = {
            "info": {'context': {"logging_context": {}},
                     'service_type': "loadbalancer",
                     'service_vendor': "haproxy",
                     },
            "config": [{
                "resource": "vip",
                "resource_data": {
                    "neutron_context": self.context,
                    "vip": self._get_vip_object()[0],
                    "old_vip": self._get_vip_object()[0]
                }}]}
        return request_data

    def get_request_data_for_create_pool(self):
        """Returns request data needed for create_pool method.

        Returns: request_data

        """

        request_data = {
            "info": {'context': {"logging_context": {}},
                     'service_type': "loadbalancer",
                     'service_vendor': "haproxy",
                     },
            "config": [{
                "resource": "pool",
                "resource_data": {
                    "neutron_context": self.context,
                    "pool": self._get_pool_object()[0],
                    "driver_name": "loadbalancer"
                }}]}
        return request_data

    def get_request_data_for_delete_pool(self):
        """Returns request data needed for delete_pool method.

        Returns: request_data

        """

        request_data = {
            "info": {'context': {"logging_context": {}},
                     'service_type': "loadbalancer",
                     'service_vendor': "haproxy",
                     },
            "config": [{
                "resource": "pool",
                "resource_data": {
                    "neutron_context": self.context,
                    "pool": self._get_pool_object()[0]
                }}]}
        return request_data

    def get_request_data_for_update_pool(self):
        """Returns request data needed for update_pool method.

        Returns: request_data

        """

        request_data = {
            "info": {'context': {"logging_context": {}},
                     'service_type': "loadbalancer",
                     'service_vendor': "haproxy",
                     },
            "config": [{
                "resource": "pool",
                "resource_data": {
                    "neutron_context": self.context,
                    "pool": self._get_pool_object()[0],
                    "old_pool": self._get_pool_object()[0]
                }}]}
        return request_data

    def get_request_data_for_member(self):
        """Returns request data needed for create_member method.

        Returns: request_data

        """

        request_data = {
            "info": {'context': {"logging_context": {}},
                     'service_type': "loadbalancer",
                     'service_vendor': "haproxy",
                     },
            "config": [{
                "resource": "member",
                "resource_data": {
                    "neutron_context": self.context,
                    "member": self._get_member_object()[0],
                }}]}
        return request_data

    def get_request_data_for_pool_hm(self):
        """Returns request data needed for create_pool_health_monitor method.

        Returns: request_data

        """

        request_data = {
            "info": {'context': {"logging_context": {}},
                     'service_type': "loadbalancer",
                     'service_vendor': "haproxy",
                     },
            "config": [{
                "resource": "pool_health_monitor",
                "resource_data": {
                    "neutron_context": self.context,
                    "health_monitor": self._get_hm_object()[0],
                    "pool_id": self._get_pool_object()[0]['id']
                }}]}
        return request_data

    def get_request_data_for_update_pool_hm(self):
        """Returns request data needed for update_pool_health_monitor method.

        Returns: request_data

        """

        request_data = {
            "info": {'context': {"logging_context": {}},
                     'service_type': "loadbalancer",
                     'service_vendor': "haproxy",
                     },
            "config": [{
                "resource": "pool_health_monitor",
                "resource_data": {
                    "neutron_context": self.context,
                    "health_monitor": self._get_hm_object()[0],
                    "pool_id": self._get_pool_object()[0]['id'],
                    "old_health_monitor": self._get_hm_object()[0]
                }}]}
        return request_data

    def get_request_data_for_update_member(self):
        """Returns request data needed for update_member method.

        Returns: request_data

        """

        request_data = {
            "info": {'context': {"logging_context": {}},
                     'service_type': "loadbalancer",
                     'service_vendor': "haproxy",
                     },
            "config": [{
                "resource": "member",
                "resource_data": {
                    "neutron_context": self.context,
                    "member": self._get_member_object()[0],
                    "old_member": self._get_member_object()[0]
                }}]}
        return request_data

    def _get_vip_object(self):
        """Returns objects that contains vip related information.

        Returns: vip

        """

        vip = [{"status": "ACTIVE",
                "protocol": "TCP",
                "description": '{"floating_ip": "192.168.100.149",'
                               '"provider_interface_mac":'
                               '"aa:bb:cc:dd:ee:ff"}',
                "address": "42.0.0.14",
                "protocol_port": 22,
                "port_id": "cfd9fcc0-c27b-478b-985e-8dd73f2c16e8",
                "id": "7a755739-1bbb-4211-9130-b6c82d9169a5",
                "status_description": None,
                "name": "lb-vip",
                "admin_state_up": True,
                "subnet_id": "b31cdafe-bdf3-4c19-b768-34d623d77d6c",
                "tenant_id": "f6b09b7a590642d8ac6de73df0ab0686",
                "connection_limit": -1,
                "pool_id": "6350c0fd-07f8-46ff-b797-62acd23760de",
                "session_persistence": None}]
        return vip

    def _get_pool_object(self):
        """Returns objects that contains pool related information.

        Returns: pool

        """

        pool = [{"status": "ACTIVE",
                 "lb_method": "ROUND_ROBIN",
                 "protocol": "TCP",
                 "description": "",
                 "health_monitors": [],
                 "members":
                 [
                     "4910851f-4af7-4592-ad04-08b508c6fa21"
                 ],
                 "status_description": None,
                 "id": "6350c0fd-07f8-46ff-b797-62acd23760de",
                 "vip_id": "7a755739-1bbb-4211-9130-b6c82d9169a5",
                 "name": "lb-pool",
                 "admin_state_up": True,
                 "subnet_id": "b31cdafe-bdf3-4c19-b768-34d623d77d6c",
                 "tenant_id": "f6b09b7a590642d8ac6de73df0ab0686",
                 "health_monitors_status": [],
                 "provider": "haproxy"}]
        return pool

    def _get_member_object(self):
        """Returns objects that contains member related information.

        Returns: member

        """

        member = [{
            "admin_state_up": True,
            "status": "ACTIVE",
            "status_description": None,
            "weight": 1,
            "address": "42.0.0.11",
            "tenant_id": "f6b09b7a590642d8ac6de73df0ab0686",
            "protocol_port": 80,
            "id": "4910851f-4af7-4592-ad04-08b508c6fa21",
            "pool_id": "6350c0fd-07f8-46ff-b797-62acd23760de"}]
        return member

    def _get_hm_object(self):
        """Returns objects that contains health_monitor related information.

        Returns: hm

        """

        hm = [{
            "admin_state_up": True,
            "tenant_id": "f6b09b7a590642d8ac6de73df0ab0686",
            "delay": 10,
            "max_retries": 3,
            "timeout": 10,
            "pools": [],
            "type": "PING",
                    "id": "c30d8a88-c719-4b93-aa64-c58efb397d86"
        }]
        return hm

    def _get_ports_object(self):
        """Returns objects that contains health_monitor related information.

        Returns: hm

        """

        ports = [{"status": "ACTIVE",
                  "name": "",
                  "allowed_address_pairs": [],
                  "admin_state_up": True,
                  "network_id": "92f423a7-f44e-4726-b453-c8a1369a3ad0",
                  "tenant_id": "5e67167662f94fd5987e12a68ea6c1d8",
                  "extra_dhcp_opts": [],
                  "binding:vnic_type": "normal",
                  "device_owner": "network:dhcp",
                  "mac_address": "fa:16:3e:01:19:11",
                  "fixed_ips": [
                      {
                          "subnet_id": "2670bdcd-1bcf-4b97-858d-ab0d621983cc",
                          "ip_address": "11.0.0.3"
                      },
                      {
                          "subnet_id": "94aee832-935b-4e23-8f90-b6a81b0195b1",
                          "ip_address": "192.168.0.2"
                      }
                  ],
                  "id": "cfd9fcc0-c27b-478b-985e-8dd73f2c16e8",
                  "security_groups": [],
                  "device_id": ("dhcpf986c817-fd54-5bae-a8e4-e473b69100d2-"
                                "92f423a7-f44e-4726-b453-c8a1369a3ad0")
                  }]
        return ports

    def _get_subnets_object(self):
        """Returns objects that contains health_monitor related information.

        Returns: hm

        """

        subnets = [{
            "name": "apic_owned_ew-consumer",
                    "enable_dhcp": True,
                    "network_id": "0ced2567-47a0-4b67-be52-0e9695e8b0e6",
                    "tenant_id": "5e67167662f94fd5987e12a68ea6c1d8",
                    "dns_nameservers": [],
                    "gateway_ip": "11.0.3.1",
                    "ipv6_ra_mode": None,
                    "allocation_pools": [
                        {
                            "start": "11.0.3.2",
                            "end": "11.0.3.254"
                        }
                    ],
            "host_routes": [],
            "ip_version": 4,
            "ipv6_address_mode": None,
            "cidr": "11.0.3.0/24",
                    "id": "ea9ff596-51bc-4381-8aff-ee9f0ef7e319"
        }]
        return subnets

""" This class provides all the data needed for event.

"""


class FakeEvent(object):

    def __init__(self):
        fo = FakeObjects()
        kwargs = {'key': 'value'}
        self.data = {
            'context': {'notification_data': {},
                        'resource': 'context_resource',
                        'agent_info': {'service_vendor': '',
                                       'service_feature': '',
                                       'context': {},
                                       'resource': ''
                                       }
                        },
            'vip': fo._get_vip_object()[0],
            'old_vip': fo._get_vip_object()[0],
            'pool': fo._get_pool_object()[0],
            'old_pool': fo._get_pool_object()[0],
            'member': fo._get_member_object()[0],
            'old_member': fo._get_member_object()[0],
            'health_monitor': fo._get_hm_object()[0],
            'old_health_monitor': fo._get_hm_object()[0],
            'pool_id': '6350c0fd-07f8-46ff-b797-62acd23760de',
            'driver_name': 'loadbalancer',
            'host': 'host',
            'kwargs': kwargs,
        }

""" This class provides assertion data for HaproxyOnVmDriverTestCase.

This class provides assertion data that is expected for mock method to
call by assert_called_with function. This assertion data is depend on
input data used to call method which uses mocking.

"""


class AssertionData(object):

    url = '192.168.100.149'
    port = '1234'
    header = {'Content-Type': 'application/json'}
    timeout = 30
    delete_vip_url = ('http://192.168.100.149:1234/backend/'
                      'bck:6350c0fd-07f8-46ff-b797-62acd23760de')

    create_vip_data = {"frnt:7a755739-1bbb-4211-9130-b6c82d9169a5":
                       {"option": {"tcplog": True},
                        "bind": "42.0.0.14:22",
                        "mode": "tcp",
                        "default_backend":
                            "bck:6350c0fd-07f8-46ff-b797-62acd23760de",
                        "provider_interface_mac": "fa:16:3e:d9:4c:33"
                        }
                       }

    create_vip_url = 'http://192.168.100.149:1234/frontend'
    create_vip_resources = 'backend/bck:6350c0fd-07f8-46ff-b797-62acd23760de'

    update_vip_data = {"option": {"tcplog": True},
                       "bind": "42.0.0.14:22",
                       "mode": "tcp",
                       "default_backend":
                           "bck:6350c0fd-07f8-46ff-b797-62acd23760de",
                       "provider_interface_mac": "fa:16:3e:d9:4c:33"
                       }

    update_vip_url = ('http://192.168.100.149:1234/frontend/frnt:'
                      '7a755739-1bbb-4211-9130-b6c82d9169a5')

    update_pool_data = {"mode": "tcp",
                        "balance": "roundrobin",
                        "option": {},
                        "timeout": {"check": "10s"},
                        "server": {
                            "srvr:4910851f-4af7-4592-ad04-08b508c6fa21":
                                ["42.0.0.11:80", "weight 1",
                                 "check inter 10s fall 3"]
                              },
                        }

    update_pool_url = ('http://192.168.100.149:1234/backend/bck:'
                       '6350c0fd-07f8-46ff-b797-62acd23760de')

    create_member_data = {"timeout": {},
                          "server":
                          {
                             "srvr:4910851f-4af7-4592-ad04-08b508c6fa21":
                             ["42.0.0.11:80", "weight 1",
                              "check inter 10s fall 3"],
                             "resource": []
                          }
                          }
    create_member_url = ('http://192.168.100.149:1234/backend/bck:'
                         '6350c0fd-07f8-46ff-b797-62acd23760de')

    delete_member_data = {"timeout": {},
                          "server": {"resource": []}
                          }

    delete_member_url = ('http://192.168.100.149:1234/backend/bck:'
                         '6350c0fd-07f8-46ff-b797-62acd23760de')

    update_member_data = create_member_data
    update_member_url = ('http://192.168.100.149:1234/backend/bck:'
                         '6350c0fd-07f8-46ff-b797-62acd23760de')

    create_hm_data = {"timeout": {"check": "10s"},
                      "server":
                      {
                         "srvr:4910851f-4af7-4592-ad04-08b508c6fa21": [],
                         "resource": []
                      }
                      }

    create_hm_url = ('http://192.168.100.149:1234/backend/bck:'
                     '6350c0fd-07f8-46ff-b797-62acd23760de')

    delete_hm_data = {"timeout": {},
                      "server":
                      {
                        "srvr:4910851f-4af7-4592-ad04-08b508c6fa21": [],
                        "resource": []
                      }
                      }

    delete_hm_url = ('http://192.168.100.149:1234/backend/bck:'
                     '6350c0fd-07f8-46ff-b797-62acd23760de')

    update_hm_data = create_hm_data
    update_hm_url = ('http://192.168.100.149:1234/backend/bck:'
                     '6350c0fd-07f8-46ff-b797-62acd23760de')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron._i18n import _LE
from neutron._i18n import _LW
from neutron.api.rpc.agentnotifiers import dhcp_rpc_agent_api
from neutron.common import constants as const
from neutron.common import exceptions as n_exc
from neutron.extensions import address_scope
from neutron.extensions import l3
from neutron.extensions import securitygroup as ext_sg
from neutron import manager
from neutron.notifiers import nova
from neutron.plugins.common import constants as pconst
from neutron import quota
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils

from gbpservice.neutron.extensions import group_policy as gp_ext
from gbpservice.neutron.extensions import servicechain as sc_ext
from gbpservice.neutron.services.grouppolicy.common import exceptions as exc

LOG = logging.getLogger(__name__)


def get_outer_transaction(transaction):
    if not transaction:
        return
    if not transaction._parent:
        return transaction
    else:
        return get_outer_transaction(transaction._parent)


BATCH_NOTIFICATIONS = False
NOTIFIER_REF = 'notifier_object_reference'
NOTIFIER_METHOD = 'notifier_method_name'
NOTIFICATION_ARGS = 'notification_args'


def _queue_notification(session, transaction_key, notifier_obj,
                        notifier_method, args):
    entry = {NOTIFIER_REF: notifier_obj, NOTIFIER_METHOD: notifier_method,
             NOTIFICATION_ARGS: args}
    if transaction_key not in session.notification_queue:
        session.notification_queue[transaction_key] = [entry]
    else:
        session.notification_queue[transaction_key].append(entry)


def send_or_queue_notification(session, transaction_key, notifier_obj,
                               notifier_method, args):
    if not transaction_key or 'subnet.delete.end' in args:
        # We make an exception for the dhcp agent notification
        # for subnet delete since the implementation for sending
        # that notification checks for the existence of the
        # subnet's network, which is not present in certain
        # cases if the subnet delete notification is queued
        # and sent after the network delete.
        getattr(notifier_obj, notifier_method)(*args)
        return

    _queue_notification(session, transaction_key, notifier_obj,
                        notifier_method, args)


def post_notifications_from_queue(session, transaction_key):
    queue = session.notification_queue[transaction_key]
    for entry in queue:
        getattr(entry[NOTIFIER_REF],
                entry[NOTIFIER_METHOD])(*entry[NOTIFICATION_ARGS])
    del session.notification_queue[transaction_key]


def discard_notifications_after_rollback(session):
    session.notification_queue.pop(session.transaction, None)


class LocalAPI(object):
    """API for interacting with the neutron Plugins directly."""

    @property
    def _nova_notifier(self):
        return nova.Notifier()

    @property
    def _core_plugin(self):
        # REVISIT(rkukura): Need initialization method after all
        # plugins are loaded to grab and store plugin.
        return manager.NeutronManager.get_plugin()

    @property
    def _l3_plugin(self):
        # REVISIT(rkukura): Need initialization method after all
        # plugins are loaded to grab and store plugin.
        plugins = manager.NeutronManager.get_service_plugins()
        l3_plugin = plugins.get(pconst.L3_ROUTER_NAT)
        if not l3_plugin:
            LOG.error(_LE("No L3 router service plugin found."))
            raise exc.GroupPolicyDeploymentError()
        return l3_plugin

    @property
    def _group_policy_plugin(self):
        # REVISIT(rkukura): Need initialization method after all
        # plugins are loaded to grab and store plugin.
        plugins = manager.NeutronManager.get_service_plugins()
        group_policy_plugin = plugins.get(pconst.GROUP_POLICY)
        if not group_policy_plugin:
            LOG.error(_LE("No GroupPolicy service plugin found."))
            raise exc.GroupPolicyDeploymentError()
        return group_policy_plugin

    @property
    def _servicechain_plugin(self):
        # REVISIT(rkukura): Need initialization method after all
        # plugins are loaded to grab and store plugin.
        plugins = manager.NeutronManager.get_service_plugins()
        servicechain_plugin = plugins.get(pconst.SERVICECHAIN)
        if not servicechain_plugin:
            LOG.error(_LE("No Servicechain service plugin found."))
            raise exc.GroupPolicyDeploymentError()
        return servicechain_plugin

    @property
    def _dhcp_agent_notifier(self):
        # REVISIT(rkukura): Need initialization method after all
        # plugins are loaded to grab and store notifier.
        if not self._cached_agent_notifier:
            agent_notifiers = getattr(self._core_plugin, 'agent_notifiers', {})
            self._cached_agent_notifier = (
                agent_notifiers.get(const.AGENT_TYPE_DHCP) or
                dhcp_rpc_agent_api.DhcpAgentNotifyAPI())
        return self._cached_agent_notifier

    def _process_notifications(self, context, action, resource, obj,
                               orig_obj=None):
        if BATCH_NOTIFICATIONS:
            outer_transaction = (get_outer_transaction(
                context._session.transaction))
        else:
            outer_transaction = None
        if orig_obj:
            args = [action, orig_obj, {resource: obj}]
        else:
            args = [action, {}, {resource: obj}]
        send_or_queue_notification(context._session,
            outer_transaction, self._nova_notifier,
            'send_network_change', args)
        # REVISIT(rkukura): Do create.end notification?
        if cfg.CONF.dhcp_agent_notification:
            action_state = ".%s.end" % action.split('_')[0]
            args = [context, {resource: obj}, resource + action_state]
            send_or_queue_notification(context._session,
                outer_transaction, self._dhcp_agent_notifier,
                'notify', args)

    def _create_resource(self, plugin, context, resource, attrs,
                         do_notify=True):
        # REVISIT(rkukura): Do create.start notification?
        # REVISIT(rkukura): Check authorization?
        reservation = None
        if plugin in [self._group_policy_plugin,
                self._servicechain_plugin]:
            reservation = quota.QUOTAS.make_reservation(
                    context, context.tenant_id, {resource: 1}, plugin)
        action = 'create_' + resource
        obj_creator = getattr(plugin, action)
        try:
            obj = obj_creator(context, {resource: attrs})
        except Exception:
            # In case of failure the plugin will always raise an
            # exception. Cancel the reservation
            with excutils.save_and_reraise_exception():
                if reservation:
                    quota.QUOTAS.cancel_reservation(
                            context, reservation.reservation_id)
        if reservation:
            quota.QUOTAS.commit_reservation(
                    context, reservation.reservation_id)
            # At this point the implicit resource creation is successfull,
            # so we should be calling:
            # resource_registry.set_resources_dirty(context)
            # to appropriately notify the quota engine. However, the above
            # call begins a new transaction and we want to avoid that.
            # Moreover, it can be safely assumed that any implicit resource
            # creation via this local_api is always in response to an
            # explicit resource creation request, and hence the above
            # method will be invoked in the API layer.
        if do_notify:
            self._process_notifications(context, action, resource, obj)
        return obj

    def _update_resource(self, plugin, context, resource, resource_id, attrs,
                         do_notify=True):
        # REVISIT(rkukura): Do update.start notification?
        # REVISIT(rkukura): Check authorization?
        obj_getter = getattr(plugin, 'get_' + resource)
        orig_obj = obj_getter(context, resource_id)
        action = 'update_' + resource
        obj_updater = getattr(plugin, action)
        obj = obj_updater(context, resource_id, {resource: attrs})
        if do_notify:
            self._process_notifications(context, action, resource, obj,
                                        orig_obj)
        return obj

    def _delete_resource(self, plugin, context, resource, resource_id,
                         do_notify=True):
        # REVISIT(rkukura): Do delete.start notification?
        # REVISIT(rkukura): Check authorization?
        obj_getter = getattr(plugin, 'get_' + resource)
        obj = obj_getter(context, resource_id)
        action = 'delete_' + resource
        obj_deleter = getattr(plugin, action)
        obj_deleter(context, resource_id)
        if do_notify:
            self._process_notifications(context, action, resource, obj)

    def _get_resource(self, plugin, context, resource, resource_id):
        obj_getter = getattr(plugin, 'get_' + resource)
        obj = obj_getter(context, resource_id)
        return obj

    def _get_resources(self, plugin, context, resource_plural, filters=None):
        obj_getter = getattr(plugin, 'get_' + resource_plural)
        obj = obj_getter(context, filters)
        return obj

    # The following methods perform the necessary subset of
    # functionality from neutron.api.v2.base.Controller.
    #
    # REVISIT(rkukura): Can we just use the WSGI Controller?  Using
    # neutronclient is also a possibility, but presents significant
    # issues to unit testing as well as overhead and failure modes.

    def _get_port(self, plugin_context, port_id):
        return self._get_resource(self._core_plugin, plugin_context, 'port',
                                  port_id)

    def _get_ports(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._core_plugin, plugin_context, 'ports',
                                   filters)

    def _create_port(self, plugin_context, attrs):
        return self._create_resource(self._core_plugin, plugin_context, 'port',
                                     attrs)

    def _update_port(self, plugin_context, port_id, attrs):
        return self._update_resource(self._core_plugin, plugin_context, 'port',
                                     port_id, attrs)

    def _delete_port(self, plugin_context, port_id):
        try:
            self._delete_resource(self._core_plugin,
                                  plugin_context, 'port', port_id)
        except n_exc.PortNotFound:
            LOG.warning(_LW('Port %s already deleted'), port_id)

    def _get_subnet(self, plugin_context, subnet_id):
        return self._get_resource(self._core_plugin, plugin_context, 'subnet',
                                  subnet_id)

    def _get_subnets(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._core_plugin, plugin_context,
                                   'subnets', filters)

    def _create_subnet(self, plugin_context, attrs):
        return self._create_resource(self._core_plugin, plugin_context,
                                     'subnet', attrs)

    def _update_subnet(self, plugin_context, subnet_id, attrs):
        return self._update_resource(self._core_plugin, plugin_context,
                                     'subnet', subnet_id, attrs)

    def _delete_subnet(self, plugin_context, subnet_id):
        try:
            self._delete_resource(self._core_plugin, plugin_context, 'subnet',
                                  subnet_id)
        except n_exc.SubnetNotFound:
            LOG.warning(_LW('Subnet %s already deleted'), subnet_id)

    def _get_network(self, plugin_context, network_id):
        return self._get_resource(self._core_plugin, plugin_context, 'network',
                                  network_id)

    def _get_networks(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(
            self._core_plugin, plugin_context, 'networks', filters)

    def _create_network(self, plugin_context, attrs):
        return self._create_resource(self._core_plugin, plugin_context,
                                     'network', attrs, True)

    def _update_network(self, plugin_context, network_id, attrs):
        return self._update_resource(self._core_plugin, plugin_context,
                                     'network', network_id, attrs)

    def _delete_network(self, plugin_context, network_id):
        try:
            self._delete_resource(self._core_plugin, plugin_context,
                                  'network', network_id)
        except n_exc.NetworkNotFound:
            LOG.warning(_LW('Network %s already deleted'), network_id)

    def _get_router(self, plugin_context, router_id):
        return self._get_resource(self._l3_plugin, plugin_context, 'router',
                                  router_id)

    def _get_routers(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._l3_plugin, plugin_context, 'routers',
                                   filters)

    def _create_router(self, plugin_context, attrs):
        return self._create_resource(self._l3_plugin, plugin_context, 'router',
                                     attrs)

    def _update_router(self, plugin_context, router_id, attrs):
        return self._update_resource(self._l3_plugin, plugin_context, 'router',
                                     router_id, attrs)

    def _add_router_interface(self, plugin_context, router_id, interface_info):
        self._l3_plugin.add_router_interface(plugin_context,
                                             router_id, interface_info)

    def _remove_router_interface(self, plugin_context, router_id,
                                 interface_info):
        # To detach Router interface either port ID or Subnet ID is mandatory
        key = 'port_id' if 'port_id' in interface_info else 'subnet_id'
        fixed_ips_filter = {key: [interface_info.get(key)]}
        filters = {'device_id': [router_id],
                   'fixed_ips': fixed_ips_filter}
        ports = self._get_ports(plugin_context, filters=filters)

        try:
            self._l3_plugin.remove_router_interface(plugin_context, router_id,
                                                    interface_info)
        except l3.RouterInterfaceNotFoundForSubnet:
            LOG.warning(_LW('Router interface already deleted for subnet %s'),
                        interface_info)
            return
        else:
            # The DHCP agent is not getting this event for the router
            # interface port delete triggered by L3 Plugin. So we are
            # sending the notification here
            if cfg.CONF.dhcp_agent_notification and ports:
                self._dhcp_agent_notifier.notify(plugin_context,
                                                 {'port': ports[0]},
                                                 'port' + '.delete.end')

    def _add_router_gw_interface(self, plugin_context, router_id, gw_info):
        return self._l3_plugin.update_router(
            plugin_context, router_id,
            {'router': {'external_gateway_info': gw_info}})

    def _remove_router_gw_interface(self, plugin_context, router_id,
                                    interface_info):
        self._l3_plugin.update_router(
            plugin_context, router_id,
            {'router': {'external_gateway_info': None}})

    def _delete_router(self, plugin_context, router_id):
        try:
            self._delete_resource(self._l3_plugin, plugin_context, 'router',
                                  router_id)
        except l3.RouterNotFound:
            LOG.warning(_LW('Router %s already deleted'), router_id)

    def _get_sg(self, plugin_context, sg_id):
        return self._get_resource(
            self._core_plugin, plugin_context, 'security_group', sg_id)

    def _get_sgs(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(
            self._core_plugin, plugin_context, 'security_groups', filters)

    def _create_sg(self, plugin_context, attrs):
        return self._create_resource(self._core_plugin, plugin_context,
                                     'security_group', attrs)

    def _update_sg(self, plugin_context, sg_id, attrs):
        return self._update_resource(self._core_plugin, plugin_context,
                                     'security_group', sg_id, attrs)

    def _delete_sg(self, plugin_context, sg_id):
        try:
            self._delete_resource(self._core_plugin, plugin_context,
                                  'security_group', sg_id)
        except ext_sg.SecurityGroupNotFound:
            LOG.warning(_LW('Security Group %s already deleted'), sg_id)

    def _get_sg_rule(self, plugin_context, sg_rule_id):
        return self._get_resource(
            self._core_plugin, plugin_context, 'security_group_rule',
            sg_rule_id)

    def _get_sg_rules(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(
            self._core_plugin, plugin_context, 'security_group_rules', filters)

    def _create_sg_rule(self, plugin_context, attrs):
        try:
            return self._create_resource(self._core_plugin, plugin_context,
                                         'security_group_rule', attrs)
        except ext_sg.SecurityGroupRuleExists as ex:
            LOG.warning(_LW('Security Group already exists %s'), ex.message)
            return

    def _update_sg_rule(self, plugin_context, sg_rule_id, attrs):
        return self._update_resource(self._core_plugin, plugin_context,
                                     'security_group_rule', sg_rule_id,
                                     attrs)

    def _delete_sg_rule(self, plugin_context, sg_rule_id):
        try:
            self._delete_resource(self._core_plugin, plugin_context,
                                  'security_group_rule', sg_rule_id)
        except ext_sg.SecurityGroupRuleNotFound:
            LOG.warning(_LW('Security Group Rule %s already deleted'),
                        sg_rule_id)

    def _get_fip(self, plugin_context, fip_id):
        return self._get_resource(
            self._l3_plugin, plugin_context, 'floatingip', fip_id)

    def _get_fips(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(
            self._l3_plugin, plugin_context, 'floatingips', filters)

    def _create_fip(self, plugin_context, attrs):
        return self._create_resource(self._l3_plugin, plugin_context,
                                     'floatingip', attrs)

    def _update_fip(self, plugin_context, fip_id, attrs):
        return self._update_resource(self._l3_plugin, plugin_context,
                                     'floatingip', fip_id, attrs)

    def _delete_fip(self, plugin_context, fip_id):
        try:
            self._delete_resource(self._l3_plugin, plugin_context,
                                  'floatingip', fip_id)
        except l3.FloatingIPNotFound:
            LOG.warning(_LW('Floating IP %s Already deleted'), fip_id)

    def _get_address_scope(self, plugin_context, address_scope_id):
        return self._get_resource(self._core_plugin, plugin_context,
                                  'address_scope', address_scope_id)

    def _get_address_scopes(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._core_plugin, plugin_context,
                                   'address_scopes', filters)

    def _create_address_scope(self, plugin_context, attrs):
        return self._create_resource(self._core_plugin, plugin_context,
                                     'address_scope', attrs)

    def _update_address_scope(self, plugin_context, address_scope_id, attrs):
        return self._update_resource(self._core_plugin, plugin_context,
                                     'address_scope', address_scope_id, attrs)

    def _delete_address_scope(self, plugin_context, address_scope_id):
        try:
            self._delete_resource(self._core_plugin, plugin_context,
                                  'address_scope', address_scope_id)
        except address_scope.AddressScopeNotFound:
            LOG.warning(_LW('Address Scope %s already deleted'),
                        address_scope_id)

    def _get_subnetpool(self, plugin_context, subnetpool_id):
        return self._get_resource(self._core_plugin, plugin_context,
                                  'subnetpool', subnetpool_id)

    def _get_subnetpools(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._core_plugin, plugin_context,
                                   'subnetpools', filters)

    def _create_subnetpool(self, plugin_context, attrs):
        return self._create_resource(self._core_plugin, plugin_context,
                                     'subnetpool', attrs)

    def _update_subnetpool(self, plugin_context, subnetpool_id, attrs):
        return self._update_resource(self._core_plugin, plugin_context,
                                     'subnetpool', subnetpool_id, attrs)

    def _delete_subnetpool(self, plugin_context, subnetpool_id):
        try:
            self._delete_resource(self._core_plugin, plugin_context,
                                  'subnetpool', subnetpool_id)
        except n_exc.SubnetpoolNotFound:
            LOG.warning(_LW('Subnetpool %s already deleted'), subnetpool_id)

    def _get_l2_policy(self, plugin_context, l2p_id):
        return self._get_resource(self._group_policy_plugin, plugin_context,
                                  'l2_policy', l2p_id)

    def _get_l2_policies(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._group_policy_plugin, plugin_context,
                                   'l2_policies', filters)

    def _create_l2_policy(self, plugin_context, attrs):
        return self._create_resource(self._group_policy_plugin, plugin_context,
                                     'l2_policy', attrs, False)

    def _update_l2_policy(self, plugin_context, l2p_id, attrs):
        return self._update_resource(self._group_policy_plugin, plugin_context,
                                     'l2_policy', l2p_id, attrs, False)

    def _delete_l2_policy(self, plugin_context, l2p_id):
        try:
            self._delete_resource(self._group_policy_plugin,
                                  plugin_context, 'l2_policy', l2p_id, False)
        except gp_ext.L2PolicyNotFound:
            LOG.warning(_LW('L2 Policy %s already deleted'), l2p_id)

    def _get_l3_policy(self, plugin_context, l3p_id):
        return self._get_resource(self._group_policy_plugin, plugin_context,
                                  'l3_policy', l3p_id)

    def _get_l3_policies(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._group_policy_plugin, plugin_context,
                                   'l3_policies', filters)

    def _create_l3_policy(self, plugin_context, attrs):
        return self._create_resource(self._group_policy_plugin, plugin_context,
                                     'l3_policy', attrs, False)

    def _update_l3_policy(self, plugin_context, l3p_id, attrs):
        return self._update_resource(self._group_policy_plugin, plugin_context,
                                     'l3_policy', l3p_id, attrs, False)

    def _delete_l3_policy(self, plugin_context, l3p_id):
        try:
            self._delete_resource(self._group_policy_plugin,
                                  plugin_context, 'l3_policy', l3p_id, False)
        except gp_ext.L3PolicyNotFound:
            LOG.warning(_LW('L3 Policy %s already deleted'), l3p_id)

    def _get_external_segment(self, plugin_context, es_id):
        return self._get_resource(self._group_policy_plugin, plugin_context,
                                  'external_segment', es_id)

    def _get_external_segments(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._group_policy_plugin, plugin_context,
                                   'external_segments', filters)

    def _create_external_segment(self, plugin_context, attrs):
        return self._create_resource(self._group_policy_plugin, plugin_context,
                                     'external_segment', attrs, False)

    def _update_external_segment(self, plugin_context, es_id, attrs):
        return self._update_resource(self._group_policy_plugin, plugin_context,
                                     'external_segment', es_id, attrs, False)

    def _delete_external_segment(self, plugin_context, es_id):
        try:
            self._delete_resource(self._group_policy_plugin, plugin_context,
                                  'external_segment', es_id, False)
        except gp_ext.ExternalSegmentNotFound:
            LOG.warning(_LW('External Segment %s already deleted'), es_id)

    def _get_external_policy(self, plugin_context, ep_id):
        return self._get_resource(self._group_policy_plugin, plugin_context,
                                  'external_policy', ep_id)

    def _get_external_policies(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._group_policy_plugin, plugin_context,
                                   'external_policies', filters)

    def _create_external_policy(self, plugin_context, attrs):
        return self._create_resource(self._group_policy_plugin, plugin_context,
                                     'external_policy', attrs, False)

    def _update_external_policy(self, plugin_context, ep_id, attrs):
        return self._update_resource(self._group_policy_plugin, plugin_context,
                                     'external_policy', ep_id, attrs, False)

    def _delete_external_policy(self, plugin_context, ep_id):
        try:
            self._delete_resource(self._group_policy_plugin, plugin_context,
                                  'external_policy', ep_id, False)
        except gp_ext.ExternalPolicyNotFound:
            LOG.warning(_LW('External Policy %s already deleted'), ep_id)

    def _get_policy_rule_set(self, plugin_context, prs_id):
        return self._get_resource(self._group_policy_plugin, plugin_context,
                                  'policy_rule_set', prs_id)

    def _get_policy_rule_sets(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._group_policy_plugin, plugin_context,
                                   'policy_rule_sets', filters)

    def _create_policy_rule_set(self, plugin_context, attrs):
        return self._create_resource(self._group_policy_plugin, plugin_context,
                                     'policy_rule_set', attrs, False)

    def _update_policy_rule_set(self, plugin_context, prs_id, attrs):
        return self._update_resource(self._group_policy_plugin, plugin_context,
                                     'policy_rule_set', prs_id, attrs, False)

    def _delete_policy_rule_set(self, plugin_context, prs_id):
        try:
            self._delete_resource(self._group_policy_plugin, plugin_context,
                                  'policy_rule_set', prs_id, False)
        except gp_ext.PolicyRuleSetNotFound:
            LOG.warning(_LW('Policy Rule Set %s already deleted'), prs_id)

    def _get_servicechain_instance(self, plugin_context, sci_id):
        return self._get_resource(self._servicechain_plugin, plugin_context,
                                  'servicechain_instance', sci_id)

    def _get_servicechain_instances(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._servicechain_plugin, plugin_context,
                                   'servicechain_instances', filters)

    def _create_servicechain_instance(self, plugin_context, attrs):
        return self._create_resource(self._servicechain_plugin, plugin_context,
                                     'servicechain_instance', attrs, False)

    def _update_servicechain_instance(self, plugin_context, sci_id, attrs):
        return self._update_resource(self._servicechain_plugin, plugin_context,
                                     'servicechain_instance', sci_id, attrs,
                                     False)

    def _delete_servicechain_instance(self, plugin_context, sci_id):
        try:
            self._delete_resource(self._servicechain_plugin, plugin_context,
                                  'servicechain_instance', sci_id, False)
        except sc_ext.ServiceChainInstanceNotFound:
            LOG.warning(_LW("servicechain %s already deleted"), sci_id)

    def _get_servicechain_spec(self, plugin_context, scs_id):
        return self._get_resource(self._servicechain_plugin, plugin_context,
                                  'servicechain_spec', scs_id)

    def _get_servicechain_specs(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._servicechain_plugin, plugin_context,
                                   'servicechain_specs', filters)

    def _create_servicechain_spec(self, plugin_context, attrs):
        return self._create_resource(self._servicechain_plugin, plugin_context,
                                     'servicechain_spec', attrs, False)

    def _update_servicechain_spec(self, plugin_context, scs_id, attrs):
        return self._update_resource(self._servicechain_plugin, plugin_context,
                                     'servicechain_spec', scs_id, attrs, False)

    def _delete_servicechain_spec(self, plugin_context, scs_id):
        try:
            self._delete_resource(self._servicechain_plugin, plugin_context,
                                  'servicechain_spec', scs_id)
        except sc_ext.ServiceChainSpecNotFound:
            LOG.warning(_LW("servicechain spec %s already deleted"), scs_id)

    def _get_policy_target(self, plugin_context, pt_id):
        return self._get_resource(self._group_policy_plugin, plugin_context,
                                  'policy_target', pt_id)

    def _get_policy_targets(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._group_policy_plugin, plugin_context,
                                   'policy_targets', filters)

    def _create_policy_target(self, plugin_context, attrs):
        return self._create_resource(self._group_policy_plugin, plugin_context,
                                     'policy_target', attrs, False)

    def _update_policy_target(self, plugin_context, pt_id, attrs):
        return self._update_resource(self._group_policy_plugin, plugin_context,
                                     'policy_target', pt_id, attrs, False)

    def _delete_policy_target(self, plugin_context, pt_id):
        try:
            self._delete_resource(self._group_policy_plugin, plugin_context,
                                  'policy_target', pt_id, False)
        except gp_ext.PolicyTargetNotFound:
            LOG.warning(_LW('Policy Rule Set %s already deleted'), pt_id)

    def _get_policy_target_group(self, plugin_context, ptg_id):
        return self._get_resource(self._group_policy_plugin, plugin_context,
                                  'policy_target_group', ptg_id)

    def _get_policy_target_groups(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._group_policy_plugin, plugin_context,
                                   'policy_target_groups', filters)

    def _create_policy_target_group(self, plugin_context, attrs):
        return self._create_resource(self._group_policy_plugin, plugin_context,
                                     'policy_target_group', attrs, False)

    def _update_policy_target_group(self, plugin_context, ptg_id, attrs):
        return self._update_resource(self._group_policy_plugin, plugin_context,
                                     'policy_target_group', ptg_id, attrs,
                                     False)

    def _delete_policy_target_group(self, plugin_context, ptg_id):
        try:
            self._delete_resource(self._group_policy_plugin, plugin_context,
                                  'policy_target_group', ptg_id)
        except sc_ext.ServiceChainSpecNotFound:
            LOG.warning(_LW("Policy Target Group %s already deleted"), ptg_id)
